# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import Counter
from typing import Any, Dict, List, Optional, Set

import onnx

import nncf
from nncf.common.graph import NNCFGraph
from nncf.common.graph.definitions import MODEL_INPUT_OP_NAME
from nncf.common.graph.definitions import MODEL_OUTPUT_OP_NAME
from nncf.common.graph.definitions import NNCFGraphNodeType
from nncf.common.graph.layer_attributes import BaseLayerAttributes
from nncf.common.graph.layer_attributes import Dtype
from nncf.common.graph.operator_metatypes import InputNoopMetatype
from nncf.common.graph.operator_metatypes import OutputNoopMetatype
from nncf.onnx.graph.metatypes.groups import CONSTANT_WEIGHT_LAYER_METATYPES
from nncf.onnx.graph.metatypes.groups import MATMUL_METATYPES
from nncf.onnx.graph.metatypes.groups import OPERATIONS_WITH_BIAS
from nncf.onnx.graph.metatypes.onnx_metatypes import ONNXGemmMetatype
from nncf.onnx.graph.metatypes.onnx_metatypes import ONNXOpMetatype
from nncf.onnx.graph.metatypes.onnx_metatypes import ONNXOpWithWeightsMetatype
from nncf.onnx.graph.metatypes.onnx_metatypes import get_metatype
from nncf.onnx.graph.metatypes.onnx_metatypes import get_tensor_edge_name
from nncf.onnx.graph.onnx_helper import get_children_node_mapping
from nncf.onnx.graph.onnx_helper import get_edge_dtype
from nncf.onnx.graph.onnx_helper import get_edge_info_mapping
from nncf.onnx.graph.onnx_helper import get_edge_shape
from nncf.onnx.graph.onnx_helper import get_input_port_id_for_node_after_input
from nncf.onnx.graph.onnx_helper import get_model_inputs
from nncf.onnx.graph.onnx_helper import get_output_port_id_for_node_before_output
from nncf.onnx.graph.onnx_helper import get_parents_node_mapping
from nncf.onnx.graph.onnx_helper import is_node_has_shared_weight


class ONNXLayerAttributes(BaseLayerAttributes):
    """
    Every NNCFNode for ONNX backend has a ONNXLayerAttributes.
    If node has weight tensor(-s), information for algorithms about weight is stored in weight_attrs.
    If node has bias tensor, information for algorithms about bias is stored in bias_attrs.
    If node has attibutes needed for algorithms, they are stored in node_attrs.
    E.g. 'transA' attirbute of Gemm node for Quantization.
    """

    def __init__(
        self,
        weight_attrs: Optional[Dict[int, Dict]] = None,
        bias_attrs: Optional[Dict[str, Any]] = None,
        node_attrs: Optional[Dict[str, Any]] = None,
    ):
        """
        :param weight_attrs: Maps input port id asocciated with weight to a weight description.
        :param bias_attrs: Maps bias tensor name asocciated with weight to a weight description.
        :param node_attrs: Maps attribute name to an attribute value.
        """
        self.weight_attrs = weight_attrs if weight_attrs is not None else {}
        self.bias_attrs = bias_attrs if bias_attrs is not None else {}
        self.node_attrs = node_attrs if node_attrs is not None else {}

    def has_weight(self) -> bool:
        return bool(self.weight_attrs)

    def has_bias(self) -> bool:
        return bool(self.bias_attrs)

    def has_node_attrs(self) -> bool:
        return bool(self.node_attrs)


def get_constant_weight_port_ids(metatype: ONNXOpMetatype) -> List[int]:
    """
    Returns port ids on which metatype must have a weight based on Operation definition.

    :param metatype: Metatype.
    :return: Port ids.
    """
    if metatype in CONSTANT_WEIGHT_LAYER_METATYPES:
        return metatype.weight_port_ids
    return []


def get_possible_weight_port_ids(metatype: ONNXOpMetatype) -> List[int]:
    """
    Returns weight port ids on which metatype could have a weight.
    Example: ONNXMatMulMetatype could have activations or weights on input port ids: 0, 1

    :param metatype: Metatype.
    :return: Port ids.
    """
    if metatype in MATMUL_METATYPES:
        return metatype.possible_weight_ports
    return []


def get_bias_tensor_port_id(metatype: ONNXOpWithWeightsMetatype) -> Optional[int]:
    """
    Returns input port id, where a bias tensor should output.

    :param node: Node, for which input port id is returned,
    :return: Input port id, where a weight bias should output or None if node can not have bias.
    """
    if metatype in OPERATIONS_WITH_BIAS:
        return metatype.bias_port_id
    return None


def _get_weight_port_ids(
    node: onnx.NodeProto,
    model: onnx.ModelProto,
    parents_node_mapping: Dict[str, onnx.NodeProto],
) -> Set[int]:
    """
    Returns all weight input ports.
    First, add constant weight port ids from metatype.
    Second, add weight port ids determined dynamically if metatype could have them.

    :param node: ONNX node.
    :param model: ONNX model.
    :param parents_node_mapping: Mapping from edge name to node which outputs this edge.
    :return: Port ids with weights.
    """
    port_ids = set()
    metatype = get_metatype(model, node)
    constant_port_ids = get_constant_weight_port_ids(metatype)
    port_ids.update(constant_port_ids)
    possible_port_ids = get_possible_weight_port_ids(metatype)
    for port_id in possible_port_ids:
        if get_tensor_edge_name(model, node, port_id, parents_node_mapping):
            port_ids.add(port_id)
    return port_ids


def _is_node_with_bias(node: onnx.NodeProto, model: onnx.ModelProto) -> bool:
    """
    Returns True if node has bias tensor, otherwise - False.

    :param node: ONNX node.
    :param model: ONNX model.
    :return: True if node has bias tensor, otherwise - False.
    """
    metatype = get_metatype(model, node)
    bias_tensor_port_id = get_bias_tensor_port_id(metatype)
    if bias_tensor_port_id is not None and len(node.input) > bias_tensor_port_id:
        return True
    return False


def _get_gemm_attrs(node: onnx.NodeProto) -> Dict[str, int]:
    """
    Returns transpose attrbiutes of GEMM node.

    :param node: GEMM node.
    :return: Trnaspose attributes.
    """
    gemm_attrs = {"transA": 0, "transB": 0}
    attribute_names = ["transA", "transB"]
    for attr in node.attribute:
        if attr.name in attribute_names:
            gemm_attrs[attr.name] = onnx.helper.get_attribute_value(attr)
    return gemm_attrs


def _get_node_attrs(node: onnx.NodeProto, model: onnx.ModelProto) -> Dict[str, Any]:
    """
    Returns node attributes.

    :param node: Node.
    :param model: ONNX model.
    :return : Node attributes.
    """
    metatype = get_metatype(model, node)
    if metatype == ONNXGemmMetatype:
        return _get_gemm_attrs(node)
    return {}


def _get_bias_attr(
    node: onnx.NodeProto,
    model: onnx.ModelProto,
    parents_node_mapping: Dict[str, onnx.NodeProto],
) -> Dict[str, str]:
    """
    Returns bias tensor attributes.

    :param node: ONNX node.
    :param model: ONNX model.
    :param parents_node_mapping: Mapping from edge name to node which outputs this edge.
    :return: Bias tensor attributes.
    """
    bias_attrs = {}
    metatype = get_metatype(model, node)
    if _is_node_with_bias(node, model):
        bias_tensor_port_id = get_bias_tensor_port_id(metatype)
        bias_edge_name = get_tensor_edge_name(model, node, bias_tensor_port_id, parents_node_mapping)
        bias_attrs["name"] = bias_edge_name
    return bias_attrs


class GraphConverter:
    """
    Builds the NNCFGraph from an ONNX model.
    """

    @staticmethod
    def _replace_empty_node_name(model: onnx.ModelProto) -> onnx.ModelProto:
        """
        Sets a unique name to every node in 'model' with empty name field.
        NNCFGraph expects every node to have a unique name.

        :param model: ONNX model.
        :return: ONNX model with filled nodes.
        """
        for i, node in enumerate(model.graph.node):
            if node.name == "":
                node.name = node.op_type + "_nncf_" + str(i)

        name_counter = Counter([node.name for node in model.graph.node])

        if max(name_counter.values()) > 1:
            raise nncf.ValidationError(
                f"Nodes {[(name, cnt) for name, cnt in name_counter.items() if cnt > 1]} "
                "(name, counts) occurred more than once. "
                "NNCF expects every node to have a unique name."
            )

        return model

    @staticmethod
    def _add_nncf_input_nodes(
        model: onnx.ModelProto,
        nncf_graph: NNCFGraph,
        edge_info_mapping: Dict[str, onnx.ValueInfoProto],
        children_node_mapping: Dict[str, List[onnx.NodeProto]],
    ) -> None:
        """
        Adds special NNCF Input nodes to NNCFGraph.
        For all the ONNX model inputs, the special NNCF Input node is placed and then corresponding edges are added.
        :param model: ONNX model.
        :param nncf_graph: NNCFGraph, in which the new nodes will be added.
        :param edge_info_mapping: Mapping from edge name to the edge info.
        :param children_node_mapping: Mapping from edge name to nodes which consume this edge as an input.
        :return: None.
        """
        for i, _input in enumerate(get_model_inputs(model)):
            input_name = _input.name
            layer_attributes = ONNXLayerAttributes()
            input_node = nncf_graph.add_nncf_node(
                node_name=MODEL_INPUT_OP_NAME + "_" + str(i),
                node_type=NNCFGraphNodeType.INPUT_NODE,
                node_metatype=InputNoopMetatype,
                layer_attributes=layer_attributes,
            )
            to_nodes = children_node_mapping[input_name]

            input_node_node_id = input_node.node_id
            edge = edge_info_mapping[input_name]
            input_shape = get_edge_shape(edge)
            onnx_dtype = get_edge_dtype(edge)
            nncf_dtype = GraphConverter.convert_onnx_dtype_to_nncf_dtype(onnx_dtype)
            output_port_id = 0

            for node in to_nodes:
                to_node_id = nncf_graph.get_node_by_name(node.name).node_id
                input_port_id = get_input_port_id_for_node_after_input(input_name, node)
                nncf_graph.add_edge_between_nncf_nodes(
                    from_node_id=input_node_node_id,
                    to_node_id=to_node_id,
                    tensor_shape=input_shape,
                    input_port_id=input_port_id,
                    output_port_id=output_port_id,
                    dtype=nncf_dtype,
                )

    @staticmethod
    def _add_nncf_output_nodes(
        model: onnx.ModelProto,
        nncf_graph: NNCFGraph,
        edge_info_mapping: Dict[str, onnx.ValueInfoProto],
        parents_node_mapping: Dict[str, onnx.NodeProto],
    ) -> None:
        """
        Adds special NNCF Output nodes to NNCFGraph.
        For all the ONNX model outputs, the special NNCF Output node is placed and then corresponding edges are added.
        :param model: ONNX model.
        :param nncf_graph: NNCFGraph, in which the new nodes will be added.
        :param edge_info_mapping: Mapping from edge name to the edge info.
        :param parents_node_mapping: Mapping from edge name to node which outputs this edge.
        :return: None.
        """
        for i, _output in enumerate(model.graph.output):
            output_name = _output.name
            if output_name in parents_node_mapping:
                layer_attributes = ONNXLayerAttributes()
                output_node = nncf_graph.add_nncf_node(
                    node_name=MODEL_OUTPUT_OP_NAME + "_" + str(i),
                    node_type=NNCFGraphNodeType.OUTPUT_NODE,
                    node_metatype=OutputNoopMetatype,
                    layer_attributes=layer_attributes,
                )

                from_node = parents_node_mapping[output_name]

                output_node_node_id = output_node.node_id
                edge = edge_info_mapping[output_name]
                output_shape = get_edge_shape(edge)
                onnx_dtype = get_edge_dtype(edge)
                nncf_dtype = GraphConverter.convert_onnx_dtype_to_nncf_dtype(onnx_dtype)
                input_port_id = 0
                from_node_id = nncf_graph.get_node_by_name(from_node.name).node_id
                output_port_id = get_output_port_id_for_node_before_output(output_name, from_node)
                nncf_graph.add_edge_between_nncf_nodes(
                    from_node_id=from_node_id,
                    to_node_id=output_node_node_id,
                    tensor_shape=output_shape,
                    input_port_id=input_port_id,
                    output_port_id=output_port_id,
                    dtype=nncf_dtype,
                )

    @staticmethod
    def convert_onnx_dtype_to_nncf_dtype(onnx_dtype: int) -> Dtype:
        """
        Converts the data type from the ONNX domain to the NNCF domain.

        :param np_dtype: ONNX data type.
        :return: NNCF data type.
        """
        return Dtype.FLOAT if onnx_dtype == int(onnx.TensorProto.FLOAT) else Dtype.INTEGER

    @staticmethod
    def create_nncf_graph(onnx_model: onnx.ModelProto) -> NNCFGraph:
        """
        Creates NNCFGraph from 'onnx_model'.
        Initially, ONNXGraph is built. All nodes from onnx_model which have valid metatype are added to NNCFGraph.
        Then, corresponding edges are added to the NNCFGraph with shape, type, output and input port ids.
        In the last step, special NNCF Input and Output nodes are added.
        :param onnx_model: ONNX model.
        :return: NNCFGraph.
        """
        onnx_model = GraphConverter._replace_empty_node_name(onnx_model)
        onnx_model = onnx.shape_inference.infer_shapes(onnx_model)
        edge_info_mapping = get_edge_info_mapping(onnx_model)
        children_node_mapping = get_children_node_mapping(onnx_model)
        parents_node_mapping = get_parents_node_mapping(onnx_model)
        nncf_graph = NNCFGraph()
        for node in onnx_model.graph.node:
            metatype = get_metatype(onnx_model, node)
            weight_port_ids = _get_weight_port_ids(node, onnx_model, parents_node_mapping)
            is_shared = None
            weight_attrs = {}
            node_attrs = _get_node_attrs(node, onnx_model)
            bias_attrs = _get_bias_attr(node, onnx_model, parents_node_mapping)
            if weight_port_ids:  # If node has weight
                weight_edge_names = []
                for weight_port_id in weight_port_ids:
                    weight_edge_name = node.input[weight_port_id]
                    weight_edge_names.append(weight_edge_name)
                    edge = edge_info_mapping[weight_edge_name]
                    weight_shape = get_edge_shape(edge)
                    weight_attrs[weight_port_id] = {"name": weight_edge_name, "shape": weight_shape}
                    if not is_shared and is_node_has_shared_weight(node, weight_port_id, children_node_mapping):
                        is_shared = True

            layer_attributes = ONNXLayerAttributes(
                weight_attrs=weight_attrs, bias_attrs=bias_attrs, node_attrs=node_attrs
            )
            nncf_graph.add_nncf_node(
                node_name=node.name,
                node_type=node.op_type,
                node_metatype=metatype,
                layer_attributes=layer_attributes,
                is_shared=is_shared,
            )

        for node in onnx_model.graph.node:
            for output_port_id, output_edge_name in enumerate(node.output):
                for consumed_node in children_node_mapping[output_edge_name]:
                    edge = edge_info_mapping.get(output_edge_name)
                    if edge is None:
                        # If the edge is None it means that the edge was not added during shape inference of ONNX model.
                        # BatchNorm exported in Training mode has unused outputs edges:
                        # mean, var, saved_mean, saved_var. NNCFGraph should not contain such edges.
                        continue
                    input_port_id = get_input_port_id_for_node_after_input(output_edge_name, consumed_node)

                    in_node_id = nncf_graph.get_node_by_name(node.name).node_id
                    output_node_id = nncf_graph.get_node_by_name(consumed_node.name).node_id
                    tensor_shape = get_edge_shape(edge)
                    onnx_dtype = get_edge_dtype(edge)
                    nncf_dtype = GraphConverter.convert_onnx_dtype_to_nncf_dtype(onnx_dtype)

                    nncf_graph.add_edge_between_nncf_nodes(
                        from_node_id=in_node_id,
                        to_node_id=output_node_id,
                        tensor_shape=tensor_shape,
                        input_port_id=input_port_id,
                        output_port_id=output_port_id,
                        dtype=nncf_dtype,
                    )

        GraphConverter._add_nncf_input_nodes(onnx_model, nncf_graph, edge_info_mapping, children_node_mapping)
        GraphConverter._add_nncf_output_nodes(onnx_model, nncf_graph, edge_info_mapping, parents_node_mapping)
        return nncf_graph



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from typing import Dict

import numpy as np
import openvino.runtime as ov

from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.graph.transformations.layout import TransformationLayout
from nncf.common.tensor_statistics.aggregator import StatisticsAggregator
from nncf.common.tensor_statistics.statistic_point import StatisticPoint
from nncf.common.tensor_statistics.statistic_point import StatisticPointsContainer
from nncf.experimental.common.tensor_statistics.collectors import MergedTensorCollector
from nncf.experimental.common.tensor_statistics.collectors import TensorCollector
from nncf.openvino.graph.node_utils import get_ov_model_reduce_node_name
from nncf.openvino.graph.node_utils import get_reducer_output_node_names
from nncf.openvino.graph.transformations.commands import OVInplaceFnInsertionCommand
from nncf.openvino.graph.transformations.commands import OVOutputInsertionCommand
from nncf.openvino.tensor import OVNNCFTensor


class OVStatisticsAggregator(StatisticsAggregator):
    def collect_statistics(self, model: ov.Model, graph: NNCFGraph) -> None:
        self._name_to_node_mapping = {op.get_friendly_name(): op for op in model.get_ops()}
        super().collect_statistics(model, graph)

    def _register_statistics(
        self, outputs: Dict[str, OVNNCFTensor], statistic_points: StatisticPointsContainer
    ) -> None:
        for _, statistic_point, tensor_collector in statistic_points.get_tensor_collectors():
            stat_node_name, port_id, _ = self._translate_to_post_layer_operation(statistic_point)
            input_info = []
            for reducer in tensor_collector.reducers:
                input_info.append(
                    [
                        hash(reducer),
                        get_reducer_output_node_names(
                            reducer.name, stat_node_name, port_id, reducer.output_port_id, reducer.inplace
                        ),
                    ]
                )
            target_inputs = TensorCollector.get_tensor_collector_inputs(outputs, input_info)
            tensor_collector.register_inputs(target_inputs)

    def _get_transformation_layout_extra_outputs(
        self, statistic_points: StatisticPointsContainer
    ) -> TransformationLayout:
        transformation_layout = TransformationLayout()
        for _, statistic_point, tensor_collector in statistic_points.get_tensor_collectors():
            for reducer in tensor_collector.reducers:
                if reducer.inplace:
                    target_node_name, target_node_port_id, _ = self._translate_to_post_layer_operation(statistic_point)
                    output_name = get_ov_model_reduce_node_name(target_node_name, reducer.name, target_node_port_id)
                    transformation_layout.register(
                        OVInplaceFnInsertionCommand(
                            statistic_point.target_point, reducer.get_inplace_fn(), reducer.output_port_id, output_name
                        )
                    )
                else:
                    transformation_layout.register(OVOutputInsertionCommand(statistic_point.target_point))
        return transformation_layout

    @staticmethod
    # TODO(dlyakhov) Move this to common part
    def _get_merged_statistic_points(
        statistic_points: StatisticPointsContainer, model: ov.Model, graph: NNCFGraph
    ) -> StatisticPointsContainer:
        merged_statistic_points = StatisticPointsContainer()
        target_type_to_tensor_collector_map = defaultdict(lambda: defaultdict(list))
        for target_node_name, _statistic_points in statistic_points.data.items():
            for statistic_point in _statistic_points:
                target_point = statistic_point.target_point
                if target_point.type in [TargetType.PRE_LAYER_OPERATION, TargetType.OPERATION_WITH_WEIGHTS]:
                    node = graph.get_node_by_name(target_node_name)
                    target_input_edge = graph.get_input_edges(node)[target_point.port_id]

                    target_type = TargetType.POST_LAYER_OPERATION
                    _target_node_name = target_input_edge.from_node.node_name
                    port_id = target_input_edge.output_port_id
                else:
                    target_type = statistic_point.target_point.type
                    _target_node_name = target_point.target_node_name
                    port_id = target_point.port_id

                # TODO: Use common target point class instead of tuple
                key = (_target_node_name, target_type, port_id)
                for tensor_collectors in statistic_point.algorithm_to_tensor_collectors.values():
                    target_type_to_tensor_collector_map[key]["collectors"].extend(tensor_collectors)
                target_type_to_tensor_collector_map[key]["target_point"].append(target_point)

        for merged_collectors_info in target_type_to_tensor_collector_map.values():
            target_point = merged_collectors_info["target_point"][0]
            collectors = merged_collectors_info["collectors"]
            merged_collector = MergedTensorCollector(collectors)
            stat_point = StatisticPoint(target_point, merged_collector, "Merged")
            merged_statistic_points.add_statistic_point(stat_point)
        return merged_statistic_points

    def _translate_to_post_layer_operation(self, statistic_point: StatisticPoint):
        target_point = statistic_point.target_point
        node_name = target_point.target_node_name
        port_id = target_point.port_id
        if target_point.type == TargetType.POST_LAYER_OPERATION:
            stat_node_name = node_name
            target_point_type = target_point.type
        elif target_point.type in [TargetType.PRE_LAYER_OPERATION, TargetType.OPERATION_WITH_WEIGHTS]:
            node = self._name_to_node_mapping[node_name]
            output = node.input_value(port_id)
            stat_node_name = output.get_node().get_friendly_name()
            port_id = output.get_index()
            target_point_type = TargetType.POST_LAYER_OPERATION
        else:
            RuntimeError(f"Unsupported target point type for statistic aggregator: {target_point.type}")
        return stat_node_name, port_id, target_point_type

    @staticmethod
    def _process_outputs(outputs: Dict[str, np.ndarray]) -> Dict[str, OVNNCFTensor]:
        return {n: OVNNCFTensor(v) for n, v in outputs.items()}



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, Optional

from nncf.common.graph.transformations.commands import TargetPoint
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.graph.transformations.commands import TransformationCommand
from nncf.common.graph.transformations.commands import TransformationType
from nncf.common.graph.transformations.layout import TransformationLayout
from nncf.tensorflow.graph.transformations.commands import TFLayer
from nncf.tensorflow.graph.transformations.commands import TFLayerPoint
from nncf.tensorflow.graph.transformations.commands import TFMultipleInsertionCommands

GRAPH_NODE_TYPES = [TargetType.LAYER, TargetType.OPERATION_WITH_WEIGHTS]


class TFTransformationLayout(TransformationLayout):
    def register(self, transformation: TransformationCommand) -> None:
        if transformation.type == TransformationType.REMOVE:
            self._transformations.append(transformation)
        elif transformation.type == TransformationType.INSERT:
            self._register_insertion_transformation(transformation)
        elif transformation.type == TransformationType.MULTI_INSERT:
            self._register_multiple_insertion_transformation(transformation)

    def _register_insertion_transformation(self, transformation: TransformationCommand) -> None:
        start_idx = self._find_transformation(
            lambda t: t.type == TransformationType.REMOVE
            and is_object_removed(t.target_point, transformation.target_point),
            reverse=True,
        )
        start_idx = 0 if start_idx is None else start_idx + 1

        idx = self._find_transformation(lambda t: t.check_command_compatibility(transformation), start_idx=start_idx)
        if idx is not None:
            self.transformations[idx] = self.transformations[idx] + transformation
            return

        idx = self._find_transformation(
            lambda t: t.type == TransformationType.MULTI_INSERT and t.check_insertion_command(transformation),
            start_idx=start_idx,
        )
        if idx is not None:
            self.transformations[idx].add_insertion_command(transformation)
            return

        idx = self._find_transformation(
            lambda t: t.type == TransformationType.INSERT
            and check_target_points(t.target_point, transformation.target_point),
            start_idx=start_idx,
        )
        if idx is not None:
            self.transformations[idx] = TFMultipleInsertionCommands(
                target_point=TFLayer(transformation.target_point.layer_name),
                check_target_points_fn=check_target_points,
                commands=[self.transformations[idx], transformation],
            )
            return

        self.transformations.append(transformation)

    def _register_multiple_insertion_transformation(self, transformation: TransformationCommand) -> None:
        start_idx = self._find_transformation(
            lambda t: t.type == TransformationType.REMOVE
            and is_object_removed(t.target_point, transformation.target_point),
            reverse=True,
        )
        start_idx = 0 if start_idx is None else start_idx + 1

        idx = self._find_transformation(lambda t: t.check_command_compatibility(transformation), start_idx=start_idx)
        if idx is not None:
            self.transformations[idx] = self.transformations[idx] + transformation
            return

        merged_transformations = []
        for t in self.transformations[start_idx:]:
            if transformation.check_insertion_command(t):
                transformation.add_insertion_command(t)
                merged_transformations.append(t)
        for t in merged_transformations:
            self.transformations.remove(t)
        self.transformations.append(transformation)

    def _find_transformation(self, condition: Callable, start_idx: int = 0, reverse: bool = False) -> Optional[int]:
        transformations_iterator = (
            reversed(list(enumerate(self.transformations[start_idx:])))
            if reverse
            else enumerate(self.transformations[start_idx:])
        )
        for idx, t in transformations_iterator:
            if condition(t):
                return idx
        return None


def check_target_points(tp0: TargetPoint, tp1: TargetPoint) -> bool:
    return (
        isinstance(tp0, TFLayerPoint)
        and isinstance(tp1, TFLayerPoint)
        and tp0.type in GRAPH_NODE_TYPES
        and tp1.type in GRAPH_NODE_TYPES
        and tp0.layer_name == tp1.layer_name
    )


def is_object_removed(removed_target: TargetPoint, command_target: TargetPoint) -> bool:
    layer_removed = (
        removed_target.type == TargetType.LAYER
        and command_target.type in GRAPH_NODE_TYPES
        and removed_target.layer_name == command_target.layer_name
    )

    operation_removed = (
        removed_target.type == TargetType.OPERATION_WITH_WEIGHTS
        and removed_target.type == command_target.type
        and removed_target.layer_name == command_target.layer_name
        and removed_target.weights_attr_name == command_target.weights_attr_name
    )

    return layer_removed or operation_removed



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from math import floor
from math import isclose
from typing import List, Set

import tensorflow as tf

import nncf
from nncf import NNCFConfig
from nncf.api.compression import CompressionLoss
from nncf.api.compression import CompressionStage
from nncf.common.accuracy_aware_training.training_loop import ADAPTIVE_COMPRESSION_CONTROLLERS
from nncf.common.graph import NNCFGraph
from nncf.common.initialization.batchnorm_adaptation import BatchnormAdaptationAlgorithm
from nncf.common.logging import nncf_logger
from nncf.common.pruning.clusterization import Cluster
from nncf.common.pruning.clusterization import Clusterization
from nncf.common.pruning.mask_propagation import MaskPropagationAlgorithm
from nncf.common.pruning.schedulers import PRUNING_SCHEDULERS
from nncf.common.pruning.schedulers import PruningScheduler
from nncf.common.pruning.shape_pruning_processor import ShapePruningProcessor
from nncf.common.pruning.statistics import FilterPruningStatistics
from nncf.common.pruning.statistics import PrunedModelStatistics
from nncf.common.pruning.statistics import PrunedModelTheoreticalBorderline
from nncf.common.pruning.utils import get_prunable_layers_in_out_channels
from nncf.common.pruning.utils import get_rounded_pruned_element_number
from nncf.common.pruning.weights_flops_calculator import WeightsFlopsCalculator
from nncf.common.schedulers import StubCompressionScheduler
from nncf.common.statistics import NNCFStatistics
from nncf.common.utils.api_marker import api
from nncf.common.utils.debug import is_debug
from nncf.config.extractors import extract_bn_adaptation_init_params
from nncf.config.schemata.defaults import PRUNING_ALL_WEIGHTS
from nncf.config.schemata.defaults import PRUNING_FILTER_IMPORTANCE
from nncf.config.schemata.defaults import PRUNING_SCHEDULE
from nncf.tensorflow.algorithm_selector import TF_COMPRESSION_ALGORITHMS
from nncf.tensorflow.graph.metatypes.common import GENERAL_CONV_LAYER_METATYPES
from nncf.tensorflow.graph.metatypes.common import LINEAR_LAYER_METATYPES
from nncf.tensorflow.graph.metatypes.matcher import get_keras_layer_metatype
from nncf.tensorflow.graph.utils import collect_wrapped_layers
from nncf.tensorflow.layers.wrapper import NNCFWrapper
from nncf.tensorflow.loss import TFZeroCompressionLoss
from nncf.tensorflow.pruning.base_algorithm import BasePruningAlgoBuilder
from nncf.tensorflow.pruning.base_algorithm import BasePruningAlgoController
from nncf.tensorflow.pruning.base_algorithm import PrunedLayerInfo
from nncf.tensorflow.pruning.filter_pruning.functions import FILTER_IMPORTANCE_FUNCTIONS
from nncf.tensorflow.pruning.filter_pruning.functions import calculate_binary_mask
from nncf.tensorflow.pruning.filter_pruning.functions import tensor_l2_normalizer
from nncf.tensorflow.pruning.operations import TF_PRUNING_OPERATOR_METATYPES
from nncf.tensorflow.pruning.operations import TFConvolutionPruningOp
from nncf.tensorflow.pruning.operations import TFElementwisePruningOp
from nncf.tensorflow.pruning.operations import TFLinearPruningOp
from nncf.tensorflow.pruning.operations import TFTransposeConvolutionPruningOp
from nncf.tensorflow.pruning.tensor_processor import TFNNCFPruningTensorProcessor
from nncf.tensorflow.pruning.utils import broadcast_filter_mask
from nncf.tensorflow.pruning.utils import collect_output_shapes
from nncf.tensorflow.pruning.utils import get_filter_axis
from nncf.tensorflow.pruning.utils import get_filters_num
from nncf.tensorflow.sparsity.magnitude.operation import BinaryMask
from nncf.tensorflow.tensor import TFNNCFTensor


@TF_COMPRESSION_ALGORITHMS.register("filter_pruning")
class FilterPruningBuilder(BasePruningAlgoBuilder):
    """
    Determines which modifications should be made to the original model in
    order to enable filter pruning during fine-tuning.
    """

    def _build_controller(self, model: tf.keras.Model):
        return FilterPruningController(
            model, self._graph, self._op_names, self._prunable_types, self._pruned_layer_groups_info, self.config
        )

    def _is_pruned_layer(self, layer: tf.keras.layers.Layer) -> bool:
        return layer.__class__.__name__ in self._prunable_types

    def _get_op_types_of_pruned_layers(self) -> List[str]:
        return [
            op_name
            for meta_op in [TFConvolutionPruningOp, TFTransposeConvolutionPruningOp, TFLinearPruningOp]
            for op_name in meta_op.get_all_op_aliases()
        ]

    def _get_types_of_grouping_ops(self) -> List[str]:
        return TFElementwisePruningOp.get_all_op_aliases()


@api()
@ADAPTIVE_COMPRESSION_CONTROLLERS.register("tf_filter_pruning")
class FilterPruningController(BasePruningAlgoController):
    """
    Controller class for the filter pruning algorithm.
    """

    def __init__(
        self,
        target_model: tf.keras.Model,
        graph: NNCFGraph,
        op_names: List[str],
        prunable_types: List[str],
        pruned_layer_groups: Clusterization[PrunedLayerInfo],
        config: NNCFConfig,
    ):
        super().__init__(target_model, op_names, prunable_types, pruned_layer_groups, config)
        self._original_graph = graph
        params = self.pruning_config.get("params", {})
        self.frozen = False
        self.pruning_quota = 0.9

        self._weights_flops_calc = WeightsFlopsCalculator(
            conv_op_metatypes=GENERAL_CONV_LAYER_METATYPES, linear_op_metatypes=LINEAR_LAYER_METATYPES
        )

        self._shape_pruning_proc = ShapePruningProcessor(
            pruning_operations_metatype=TF_PRUNING_OPERATOR_METATYPES, prunable_types=prunable_types
        )

        self._pruning_quotas = {}
        self._next_nodes = {}
        self._output_shapes = {}
        _, output_channels = get_prunable_layers_in_out_channels(self._original_graph)
        self._init_pruned_layers_params(output_channels)

        self.full_flops, self.full_params_num = self._weights_flops_calc.count_flops_and_weights(
            graph, self._output_shapes
        )
        self.full_filters_num = self._weights_flops_calc.count_filters_num(graph, output_channels)

        self.current_flops, self.current_params_num = self.full_flops, self.full_params_num
        self.current_filters_num = self.full_filters_num

        self._pruned_layers_num = len(self._pruned_layer_groups_info.get_all_nodes())
        self._prunable_layers_num = len(self._original_graph.get_nodes_by_types(self._prunable_types))
        (
            self._min_possible_flops,
            self._min_possible_params,
        ) = self._calculate_flops_and_weights_in_uniformly_pruned_model(1.0)

        self._weights_normalizer = tensor_l2_normalizer  # for all weights in common case
        self._filter_importance = FILTER_IMPORTANCE_FUNCTIONS.get(
            params.get("filter_importance", PRUNING_FILTER_IMPORTANCE)
        )
        self.all_weights = params.get("all_weights", PRUNING_ALL_WEIGHTS)
        scheduler_cls = PRUNING_SCHEDULERS.get(params.get("schedule", PRUNING_SCHEDULE))
        self._scheduler = scheduler_cls(self, params)
        self._bn_adaptation = None
        self.set_pruning_level(self.pruning_init)
        self._loss = TFZeroCompressionLoss()

    @property
    def scheduler(self) -> PruningScheduler:
        return self._scheduler

    @property
    def loss(self) -> CompressionLoss:
        return self._loss

    def compression_stage(self) -> CompressionStage:
        target_pruning_level = self.scheduler.target_level
        actual_pruning_level = self.pruning_level
        if actual_pruning_level == 0:
            return CompressionStage.UNCOMPRESSED
        if (
            isclose(actual_pruning_level, target_pruning_level, abs_tol=1e-5)
            or actual_pruning_level > target_pruning_level
        ):
            return CompressionStage.FULLY_COMPRESSED
        return CompressionStage.PARTIALLY_COMPRESSED

    @property
    def compression_rate(self) -> float:
        if self.prune_flops:
            return 1 - self.current_flops / self.full_flops
        return self.pruning_level

    @compression_rate.setter
    def compression_rate(self, compression_rate: float) -> None:
        is_pruning_controller_frozen = self.frozen
        self.freeze(False)
        self.set_pruning_level(compression_rate)
        self.freeze(is_pruning_controller_frozen)

    def disable_scheduler(self):
        self._scheduler = StubCompressionScheduler()
        self._scheduler.current_pruning_level = 0.0

    def statistics(self, quickly_collected_only: bool = False) -> NNCFStatistics:
        if not quickly_collected_only and is_debug():
            stats = PrunedModelTheoreticalBorderline(
                self._pruned_layers_num,
                self._prunable_layers_num,
                self._min_possible_flops,
                self._min_possible_params,
                self.full_flops,
                self.full_params_num,
            )

            nncf_logger.debug(stats.to_str())

        pruned_layers_summary = self._calculate_pruned_layers_summary()
        self._update_benchmark_statistics()
        model_statistics = PrunedModelStatistics(
            self.full_flops,
            self.current_flops,
            self.full_params_num,
            self.current_params_num,
            self.full_filters_num,
            self.current_filters_num,
            pruned_layers_summary,
        )

        stats = FilterPruningStatistics(
            model_statistics, self.scheduler.current_pruning_level, self.scheduler.target_level, self.prune_flops
        )

        nncf_stats = NNCFStatistics()
        nncf_stats.register("filter_pruning", stats)
        return nncf_stats

    def freeze(self, freeze: bool = True):
        self.frozen = freeze

    def set_pruning_level(self, pruning_level: float, run_batchnorm_adaptation: bool = False):
        """
        Setup pruning masks in accordance to provided pruning level.

        :param pruning_level: Pruning level to be set.
        :param run_batchnorm_adaptation: Whether to run batchnorm adaptation after setting the pruning level.
        """
        # Pruning level from scheduler can be percentage of params that should be pruned
        self.pruning_level = pruning_level
        if not self.frozen:
            nncf_logger.info("Computing filter importance scores and binary masks...")
            if self.all_weights:
                if self.prune_flops:
                    self._set_binary_masks_for_pruned_modules_globally_by_flops_target(pruning_level)
                else:
                    self._set_binary_masks_for_pruned_layers_globally(pruning_level)
            else:
                if self.prune_flops:
                    # Looking for a layerwise pruning level needed for the required flops pruning level
                    pruning_level = self._find_uniform_pruning_level_for_target_flops(pruning_level)
                self._set_binary_masks_for_pruned_layers_groupwise(pruning_level)

        if run_batchnorm_adaptation:
            self._run_batchnorm_adaptation()

    @property
    def maximal_compression_rate(self) -> float:
        if self.prune_flops:
            return 1 - self._min_possible_flops / max(self.full_flops, 1)
        return 1.0

    def _init_pruned_layers_params(self, output_channels):
        # 1. Collect nodes output shapes
        self._output_shapes = collect_output_shapes(self._model, self._original_graph)

        # 2. Initialize next_nodes for each pruning cluster
        self._next_nodes = self._shape_pruning_proc.get_next_nodes(self._original_graph, self._pruned_layer_groups_info)

        # 3. Initialize pruning quotas
        for cluster in self._pruned_layer_groups_info.get_all_clusters():
            self._pruning_quotas[cluster.id] = floor(
                output_channels[cluster.elements[0].node_name] * self.pruning_quota
            )

    def _set_binary_masks_for_pruned_layers_groupwise(self, pruning_level: float):
        nncf_logger.debug("Setting new binary masks for pruned layers.")
        wrapped_layers = collect_wrapped_layers(self._model)

        # 0. Removing masks at the elements of the NNCFGraph
        for node in self._original_graph.topological_sort():
            node.attributes.pop("output_mask", None)

        # 1. Calculate masks
        for group in self._pruned_layer_groups_info.get_all_clusters():
            # a. Calculate the cumulative importance for all filters in the group
            cumulative_filters_importance = self._calculate_filters_importance_in_group(group)
            filters_num = len(cumulative_filters_importance)

            # b. Calculate threshold
            num_of_sparse_elems = get_rounded_pruned_element_number(
                cumulative_filters_importance.shape[0], pruning_level
            )
            threshold = sorted(cumulative_filters_importance)[min(num_of_sparse_elems, filters_num - 1)]

            # c. Initialize masks
            filter_mask = calculate_binary_mask(cumulative_filters_importance, threshold)
            for node in group.elements:
                nncf_node = self._original_graph.get_node_by_id(node.nncf_node_id)
                nncf_node.attributes["output_mask"] = TFNNCFTensor(filter_mask)

        # 2. Propagating masks across the graph
        mask_propagator = MaskPropagationAlgorithm(
            self._original_graph, TF_PRUNING_OPERATOR_METATYPES, TFNNCFPruningTensorProcessor
        )
        mask_propagator.mask_propagation()

        # 3. Apply masks to the model
        nncf_sorted_nodes = self._original_graph.topological_sort()
        for layer in wrapped_layers:
            nncf_node = [n for n in nncf_sorted_nodes if layer.name == n.layer_name][0]
            if nncf_node.attributes["output_mask"] is not None:
                self._set_operation_masks([layer], nncf_node.attributes["output_mask"].tensor)

        # Calculate actual flops and weights number with new masks
        self._update_benchmark_statistics()

    def _set_binary_masks_for_pruned_layers_globally(self, pruning_level: float):
        """
        Sets the binary mask values for layer groups according to the global pruning level.
        Filter importance scores in each group are merged into a single global list and a
        threshold value separating the pruning_level proportion of the least important filters
        in the model is calculated. Filters are pruned globally according to the threshold value.
        """
        nncf_logger.debug("Setting new binary masks for all pruned modules together.")
        filter_importances = {}
        wrapped_layers = collect_wrapped_layers(self._model)

        # 0. Remove masks at the elements of the NNCFGraph
        for node in self._original_graph.topological_sort():
            node.attributes.pop("output_mask", None)

        # 1. Calculate masks
        # a. Calculate importances for all groups of filters
        for group in self._pruned_layer_groups_info.get_all_clusters():
            cumulative_filters_importance = self._calculate_filters_importance_in_group(group)
            filter_importances[group.id] = cumulative_filters_importance

        # b. Calculate one threshold for all weights
        importances = tf.concat(list(filter_importances.values()), 0)
        threshold = sorted(importances)[int(pruning_level * importances.shape[0])]

        # c. Initialize masks
        for group in self._pruned_layer_groups_info.get_all_clusters():
            filter_mask = calculate_binary_mask(filter_importances[group.id], threshold)
            for node in group.elements:
                nncf_node = self._original_graph.get_node_by_id(node.nncf_node_id)
                nncf_node.attributes["output_mask"] = TFNNCFTensor(filter_mask)

        # 2. Propagate masks across the graph
        mask_propagator = MaskPropagationAlgorithm(
            self._original_graph, TF_PRUNING_OPERATOR_METATYPES, TFNNCFPruningTensorProcessor
        )
        mask_propagator.mask_propagation()

        # 3. Apply masks to the model
        nncf_sorted_nodes = self._original_graph.topological_sort()
        for layer in wrapped_layers:
            nncf_node = [n for n in nncf_sorted_nodes if layer.name == n.layer_name][0]
            if nncf_node.attributes["output_mask"] is not None:
                self._set_operation_masks([layer], nncf_node.attributes["output_mask"].tensor)

        # Calculate actual flops with new masks
        self._update_benchmark_statistics()

    def _set_binary_masks_for_pruned_modules_globally_by_flops_target(self, target_flops_pruning_level: float):
        """
        Prunes least important filters one-by-one until target FLOPs pruning level is achieved.
        Filters are sorted by filter importance score.
        """
        nncf_logger.debug("Setting new binary masks for pruned layers.")
        target_flops = self.full_flops * (1 - target_flops_pruning_level)
        wrapped_layers = collect_wrapped_layers(self._model)
        masks = {}

        nncf_sorted_nodes = self._original_graph.topological_sort()
        for layer in wrapped_layers:
            nncf_node = [n for n in nncf_sorted_nodes if layer.name == n.layer_name][0]
            nncf_node.attributes["output_mask"] = TFNNCFTensor(tf.ones(get_filters_num(layer)))

        # 1. Calculate importances for all groups of filters. Initialize masks.
        filter_importances = []
        group_indexes = []
        filter_indexes = []
        for group in self._pruned_layer_groups_info.get_all_clusters():
            cumulative_filters_importance = self._calculate_filters_importance_in_group(group)
            filter_importances.extend(cumulative_filters_importance)
            filters_num = len(cumulative_filters_importance)
            group_indexes.extend([group.id] * filters_num)
            filter_indexes.extend(range(filters_num))
            masks[group.id] = tf.ones(filters_num)

        # 2.
        tmp_in_channels, tmp_out_channels = get_prunable_layers_in_out_channels(self._original_graph)
        sorted_importances = sorted(zip(filter_importances, group_indexes, filter_indexes), key=lambda x: x[0])
        for _, group_id, filter_index in sorted_importances:
            if self._pruning_quotas[group_id] == 0:
                continue
            masks[group_id] = tf.tensor_scatter_nd_update(masks[group_id], [[filter_index]], [0])
            self._pruning_quotas[group_id] -= 1

            cluster = self._pruned_layer_groups_info.get_cluster_by_id(group_id)
            # Update input/output shapes of pruned elements
            self._shape_pruning_proc.prune_cluster_shapes(
                cluster=cluster,
                pruned_elems=1,
                pruning_groups_next_nodes=self._next_nodes,
                input_channels=tmp_in_channels,
                output_channels=tmp_out_channels,
            )

            flops, params_num = self._weights_flops_calc.count_flops_and_weights(
                graph=self._original_graph,
                output_shapes=self._output_shapes,
                input_channels=tmp_in_channels,
                output_channels=tmp_out_channels,
            )
            if flops <= target_flops:
                # 3. Add masks to the graph and propagate them
                for group in self._pruned_layer_groups_info.get_all_clusters():
                    for node in group.elements:
                        nncf_node = self._original_graph.get_node_by_id(node.nncf_node_id)
                        nncf_node.attributes["output_mask"] = TFNNCFTensor(masks[group.id])

                mask_propagator = MaskPropagationAlgorithm(
                    self._original_graph, TF_PRUNING_OPERATOR_METATYPES, TFNNCFPruningTensorProcessor
                )
                mask_propagator.mask_propagation()

                # 4. Set binary masks to the model
                self.current_flops = flops
                self.current_params_num = params_num
                nncf_sorted_nodes = self._original_graph.topological_sort()
                for layer in wrapped_layers:
                    nncf_node = [n for n in nncf_sorted_nodes if layer.name == n.layer_name][0]
                    if nncf_node.attributes["output_mask"] is not None:
                        self._set_operation_masks([layer], nncf_node.attributes["output_mask"].tensor)
                return
        raise nncf.InternalError(f"Unable to prune model to required flops pruning level: {target_flops_pruning_level}")

    def _set_operation_masks(self, layers: List[NNCFWrapper], filter_mask):
        for layer in layers:
            for weight_attr, ops in layer.weights_attr_ops.items():
                weight_shape = layer.layer_weights[weight_attr].shape
                for op_name, op in ops.items():
                    if isinstance(op, BinaryMask):
                        filter_axis = get_filter_axis(layer, weight_attr)
                        broadcasted_mask = broadcast_filter_mask(filter_mask, weight_shape, filter_axis)
                        layer.ops_weights[op_name]["mask"].assign(broadcasted_mask)

    def _find_uniform_pruning_level_for_target_flops(self, target_flops_pruning_level):
        error = 0.01
        target_flops = self.full_flops * (1 - target_flops_pruning_level)
        left, right = 0.0, 1.0
        while abs(right - left) > error:
            middle = (left + right) / 2
            flops, params_num = self._calculate_flops_and_weights_in_uniformly_pruned_model(middle)
            if flops < target_flops:
                right = middle
            else:
                left = middle
        flops, params_num = self._calculate_flops_and_weights_in_uniformly_pruned_model(right)
        if flops <= target_flops:
            self.current_flops = flops
            self.current_params_num = params_num
            return right
        raise nncf.ParameterNotSupportedError(
            f"Unable to prune the model to get the required pruning level in flops = {target_flops_pruning_level}"
        )

    def _calculate_flops_and_weights_in_uniformly_pruned_model(self, pruning_level):
        (
            tmp_in_channels,
            tmp_out_channels,
        ) = self._shape_pruning_proc.calculate_in_out_channels_in_uniformly_pruned_model(
            self._original_graph,
            pruning_groups=self._pruned_layer_groups_info,
            pruning_groups_next_nodes=self._next_nodes,
            pruning_level=pruning_level,
        )

        return self._weights_flops_calc.count_flops_and_weights(
            graph=self._original_graph,
            output_shapes=self._output_shapes,
            input_channels=tmp_in_channels,
            output_channels=tmp_out_channels,
        )

    def _calculate_filters_importance_in_group(self, group: Cluster[PrunedLayerInfo]):
        """
        Calculates cumulative filters importance in the group.
        :param group: Nodes cluster
        :return a list of filter importance scores
        """
        group_layers = [self._model.get_layer(node.layer_name) for node in group.elements]
        group_filters_num = tf.constant([get_filters_num(layer) for layer in group_layers])
        filters_num = group_filters_num[0]
        assert tf.reduce_all(group_filters_num == filters_num)

        cumulative_filters_importance = tf.zeros(filters_num)
        # Calculate cumulative importance for all filters in this group
        shared_nodes: Set[str] = set()
        for minfo in group.elements:
            layer_name = minfo.layer_name
            if layer_name in shared_nodes:
                continue
            nncf_node = self._original_graph.get_node_by_id(minfo.nncf_node_id)
            if nncf_node.is_shared():
                shared_nodes.add(layer_name)
            filters_importance = self._layer_filter_importance(self._model.get_layer(layer_name))
            cumulative_filters_importance += filters_importance

        return cumulative_filters_importance

    def _update_benchmark_statistics(self):
        tmp_in_channels, tmp_out_channels = self._shape_pruning_proc.calculate_in_out_channels_by_masks(
            graph=self._original_graph,
            pruning_groups=self._pruned_layer_groups_info,
            pruning_groups_next_nodes=self._next_nodes,
            num_of_sparse_elements_by_node=self._calculate_num_of_sparse_elements_by_node(),
        )

        self.current_filters_num = self._weights_flops_calc.count_filters_num(
            graph=self._original_graph, output_channels=tmp_out_channels
        )

        self.current_flops, self.current_params_num = self._weights_flops_calc.count_flops_and_weights(
            graph=self._original_graph,
            output_shapes=self._output_shapes,
            input_channels=tmp_in_channels,
            output_channels=tmp_out_channels,
        )

    def _layer_filter_importance(self, layer: NNCFWrapper):
        layer_metatype = get_keras_layer_metatype(layer)
        if len(layer_metatype.weight_definitions) != 1:
            raise nncf.InternalError(
                f"The layer {layer.layer.name} does not support by the pruning "
                f"algorithm because it contains several weight attributes."
            )
        weight_attr = layer_metatype.weight_definitions[0].weight_attr_name
        weight = layer.layer_weights[weight_attr]
        if self.all_weights:
            weight = self._weights_normalizer(weight)
        target_weight_dim_for_compression = get_filter_axis(layer, weight_attr)
        filters_importance = self._filter_importance(weight, target_weight_dim_for_compression)
        return filters_importance

    def _run_batchnorm_adaptation(self):
        if self._bn_adaptation is None:
            self._bn_adaptation = BatchnormAdaptationAlgorithm(
                **extract_bn_adaptation_init_params(self.config, "filter_pruning")
            )
        self._bn_adaptation.run(self.model)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from copy import deepcopy
from typing import List

import torch
import torch.distributed as dist

from nncf import NNCFConfig
from nncf.api.compression import CompressionStage
from nncf.common.accuracy_aware_training.training_loop import ADAPTIVE_COMPRESSION_CONTROLLERS
from nncf.common.graph import NNCFNode
from nncf.common.schedulers import StubCompressionScheduler
from nncf.common.sparsity.schedulers import SPARSITY_SCHEDULERS
from nncf.common.sparsity.statistics import RBSparsityStatistics
from nncf.common.statistics import NNCFStatistics
from nncf.common.utils.api_marker import api
from nncf.config.extractors import extract_algo_specific_config
from nncf.config.schemata.defaults import SPARSITY_INIT
from nncf.config.schemata.defaults import SPARSITY_LEVEL_SETTING_MODE
from nncf.torch.algo_selector import PT_COMPRESSION_ALGORITHMS
from nncf.torch.compression_method_api import PTCompressionAlgorithmController
from nncf.torch.nncf_network import NNCFNetwork
from nncf.torch.sparsity.base_algo import BaseSparsityAlgoBuilder
from nncf.torch.sparsity.base_algo import BaseSparsityAlgoController
from nncf.torch.sparsity.base_algo import SparseModuleInfo
from nncf.torch.sparsity.collector import PTSparseModelStatisticsCollector
from nncf.torch.sparsity.rb.layers import RBSparsifyingWeight
from nncf.torch.sparsity.rb.loss import SparseLoss
from nncf.torch.sparsity.rb.loss import SparseLossForPerLayerSparsity
from nncf.torch.utils import get_model_device
from nncf.torch.utils import get_world_size


@PT_COMPRESSION_ALGORITHMS.register("rb_sparsity")
class RBSparsityBuilder(BaseSparsityAlgoBuilder):
    def create_weight_sparsifying_operation(self, target_module_node: NNCFNode, compression_lr_multiplier: float):
        return RBSparsifyingWeight(
            target_module_node.layer_attributes.get_weight_shape(),
            frozen=False,
            compression_lr_multiplier=compression_lr_multiplier,
        )

    def _build_controller(self, model: NNCFNetwork) -> PTCompressionAlgorithmController:
        return RBSparsityController(model, self._sparsified_module_info, self.config)


@api()
@ADAPTIVE_COMPRESSION_CONTROLLERS.register("pt_rb_sparsity")
class RBSparsityController(BaseSparsityAlgoController):
    """
    Controller for the regularization-based (RB) sparsity algorithm in PT.
    """

    def __init__(self, target_model: NNCFNetwork, sparsified_module_info: List[SparseModuleInfo], config: NNCFConfig):
        super().__init__(target_model, sparsified_module_info)
        algo_config = extract_algo_specific_config(config, "rb_sparsity")
        params = deepcopy(algo_config.get("params", {}))

        self._distributed = False
        self._mode = params.get("sparsity_level_setting_mode", SPARSITY_LEVEL_SETTING_MODE)
        self._check_sparsity_masks = params.get("check_sparsity_masks", False)

        sparsify_operations = [m.operand for m in self.sparsified_module_info]
        if self._mode == "local":
            self._loss = SparseLossForPerLayerSparsity(sparsify_operations)
            self._scheduler = StubCompressionScheduler()
        else:
            self._loss = SparseLoss(sparsify_operations)

            sparsity_init = algo_config.get("sparsity_init", SPARSITY_INIT)
            params["sparsity_init"] = sparsity_init
            scheduler_cls = SPARSITY_SCHEDULERS.get(params.get("schedule", "exponential"))
            self._scheduler = scheduler_cls(self, params)
            self.set_sparsity_level(sparsity_init)

    def set_sparsity_level(self, sparsity_level, target_sparsified_module_info: SparseModuleInfo = None):
        if target_sparsified_module_info is None:
            self._loss.set_target_sparsity_loss(sparsity_level)
        else:
            sparse_op = target_sparsified_module_info.operand
            self._loss.set_target_sparsity_loss(sparsity_level, sparse_op)

    @property
    def current_sparsity_level(self) -> float:
        return self._loss.current_sparsity

    def compression_stage(self) -> CompressionStage:
        if self._mode == "local":
            return CompressionStage.FULLY_COMPRESSED

        if self.scheduler.current_sparsity_level == 0:
            return CompressionStage.UNCOMPRESSED
        if self.scheduler.current_sparsity_level >= self.scheduler.target_level:
            return CompressionStage.FULLY_COMPRESSED
        return CompressionStage.PARTIALLY_COMPRESSED

    def freeze(self):
        self._loss.disable()

    def distributed(self):
        if not dist.is_initialized():
            raise KeyError(
                "Could not set distributed mode for the compression algorithm "
                "because the default process group has not been initialized."
            )

        if "cuda" in get_model_device(self._model).type:
            state = torch.cuda.get_rng_state()
            if dist.get_backend() == dist.Backend.NCCL:
                state = state.cuda()
            torch.distributed.broadcast(state, src=0)
            torch.cuda.set_rng_state(state.cpu())
        else:
            state = torch.get_rng_state()
            torch.distributed.broadcast(state, src=0)
            torch.set_rng_state(state)

        self._distributed = True

    def _check_distributed_masks(self):
        if not self._distributed or get_world_size() == 1:
            return 1

        nvalues = 0
        ncor_values = 0
        eps = 1e-4
        for minfo in self.sparsified_module_info:
            mask = minfo.operand.mask

            mask_list = [torch.empty_like(mask) for _ in range(get_world_size())]
            # nccl does not support gather, send, recv operations
            dist.all_gather(mask_list, mask)

            for i in range(1, len(mask_list)):
                rel_error = (mask_list[0] - mask_list[i]) / mask_list[0]
                ncor_values = ncor_values + (rel_error.abs() < eps).sum(dtype=mask.dtype)
                nvalues = nvalues + mask_list[i].numel()

        return ncor_values / nvalues

    def statistics(self, quickly_collected_only=False) -> NNCFStatistics:
        collector = PTSparseModelStatisticsCollector(self.model, self.sparsified_module_info)
        model_statistics = collector.collect()

        target_sparsity_level = self.scheduler.current_sparsity_level if self._mode == "global" else None

        mean_sparse_prob = 1.0 - self.loss.mean_sparse_prob

        stats = RBSparsityStatistics(model_statistics, target_sparsity_level, mean_sparse_prob)

        nncf_stats = NNCFStatistics()
        nncf_stats.register("rb_sparsity", stats)
        return nncf_stats

    @property
    def compression_rate(self):
        return self._loss.target_sparsity_rate

    @compression_rate.setter
    def compression_rate(self, sparsity_level: float):
        self.set_sparsity_level(sparsity_level)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, Optional

from nncf.common.logging import nncf_logger
from nncf.common.schedulers import BaseCompressionScheduler
from nncf.common.schedulers import ExponentialDecaySchedule
from nncf.common.schedulers import MultiStepSchedule
from nncf.common.schedulers import PolynomialDecaySchedule
from nncf.common.sparsity.controller import SparsityController
from nncf.common.utils.registry import Registry
from nncf.config.schemata.defaults import SPARSITY_FREEZE_EPOCH
from nncf.config.schemata.defaults import SPARSITY_INIT
from nncf.config.schemata.defaults import SPARSITY_MULTISTEP_SPARSITY_LEVELS
from nncf.config.schemata.defaults import SPARSITY_MULTISTEP_STEPS
from nncf.config.schemata.defaults import SPARSITY_SCHEDULER_CONCAVE
from nncf.config.schemata.defaults import SPARSITY_SCHEDULER_PATIENCE
from nncf.config.schemata.defaults import SPARSITY_SCHEDULER_POWER
from nncf.config.schemata.defaults import SPARSITY_SCHEDULER_UPDATE_PER_OPTIMIZER_STEP
from nncf.config.schemata.defaults import SPARSITY_TARGET
from nncf.config.schemata.defaults import SPARSITY_TARGET_EPOCH

SPARSITY_SCHEDULERS: Registry = Registry("sparsity_schedulers")


class SparsityScheduler(BaseCompressionScheduler):
    """
    This is the class from which all sparsity schedulers inherit.

    A sparsity scheduler is an object which specifies the sparsity
    level at each training epoch or each training step. It involves a
    scheduling algorithm, defined in the `_calculate_sparsity_level()`
    method and a state (some parameters required for current sparsity
    level calculation) defined in the `__init__()` method.

    :param initial_level: Sparsity level which already has been
        applied to the model. It is the level at which the schedule begins.
    :param target_level: Sparsity level at which the schedule ends.
    :param target_epoch: Zero-based index of the epoch from which the
        sparsity level of the model will be equal to the `target_level`.
    :param freeze_epoch: Zero-based index of the epoch from which the sparsity
        mask will be frozen and will not be trained.
    """

    def __init__(self, controller: SparsityController, params: Dict[str, Any]) -> None:
        """
        Initializes the internal state of the sparsity scheduler.

        :param controller: Sparsity algorithm controller.
        :param params: Parameters of the scheduler.
        """
        super().__init__()
        self._controller = controller
        self.initial_level: float = params.get("sparsity_init", SPARSITY_INIT)
        self.target_level: float = params.get("sparsity_target", SPARSITY_TARGET)
        self.target_epoch = params.get("sparsity_target_epoch", SPARSITY_TARGET_EPOCH)
        self.freeze_epoch = params.get("sparsity_freeze_epoch", SPARSITY_FREEZE_EPOCH)

    def _calculate_sparsity_level(self) -> float:
        """
        Calculates a sparsity level that should be applied to the weights
        for the `current_epoch` or for step in the `current_epoch`.

        :return: Sparsity level that should be applied to the weights
            for the `current_epoch` or for step in the `current_epoch`.
        """
        raise NotImplementedError("SparsityScheduler implementation must override _calculate_sparsity_level method.")

    def _update_sparsity_level(self) -> None:
        """
        Calculates the current sparsity level and updates the internal
        state of the `controller`.
        """
        if self.current_epoch >= self.freeze_epoch:
            self._controller.freeze()
        self._controller.set_sparsity_level(self._calculate_sparsity_level())

    @property
    def current_sparsity_level(self) -> float:
        """
        Returns sparsity level for the `current_epoch` or for step
        in the `current_epoch`.

        :return: Current sparsity level.
        """
        if self._current_epoch == -1:
            return self.initial_level
        return self._calculate_sparsity_level()


@SPARSITY_SCHEDULERS.register("polynomial")
class PolynomialSparsityScheduler(SparsityScheduler):
    """
    Sparsity scheduler with a polynomial decay schedule.

    Two ways are available for calculations of the sparsity:
        - per epoch
        - per step
    Parameters `update_per_optimizer_step` and `steps_per_epoch`
    should be provided in config for the per step calculation.
    If `update_per_optimizer_step` was only provided then scheduler
    will use first epoch to calculate `steps_per_epoch`
    parameter. In this case, `current_epoch` and `current_step` will
    not be updated on this epoch. The scheduler will start calculation
    after `steps_per_epoch` will be calculated.
    """

    def __init__(self, controller: SparsityController, params: Dict[str, Any]):
        """
        Initializes a sparsity scheduler with a polynomial decay schedule.

        :param controller: Sparsity algorithm controller.
        :param params: Parameters of the scheduler.
        """
        super().__init__(controller, params)
        self.schedule = PolynomialDecaySchedule(
            self.initial_level,
            self.target_level,
            self.target_epoch,
            params.get("power", SPARSITY_SCHEDULER_POWER),
            params.get("concave", SPARSITY_SCHEDULER_CONCAVE),
        )
        self._steps_in_current_epoch = 0
        self._update_per_optimizer_step = params.get(
            "update_per_optimizer_step", SPARSITY_SCHEDULER_UPDATE_PER_OPTIMIZER_STEP
        )
        self._steps_per_epoch = params.get("steps_per_epoch")
        self._should_skip = False

    def step(self, next_step: Optional[int] = None) -> None:
        self._steps_in_current_epoch += 1
        if self._should_skip:
            return

        super().step(next_step)
        if self._update_per_optimizer_step:
            self._update_sparsity_level()

    def epoch_step(self, next_epoch: Optional[int] = None) -> None:
        self._maybe_should_skip()
        if self._should_skip:
            return

        self._steps_in_current_epoch = 0

        super().epoch_step(next_epoch)
        if not self._update_per_optimizer_step:
            self._update_sparsity_level()

    def _calculate_sparsity_level(self) -> float:
        local_step = max(self._steps_in_current_epoch - 1, 0)
        return self.schedule(self.current_epoch, local_step, self._steps_per_epoch)

    def load_state(self, state: Dict[str, Any]) -> None:
        super().load_state(state)
        if self._update_per_optimizer_step:
            self._steps_per_epoch = state["_steps_per_epoch"]

    def get_state(self) -> Dict[str, Any]:
        state = super().get_state()
        if self._update_per_optimizer_step:
            state["_steps_per_epoch"] = self._steps_per_epoch
        return state

    def _maybe_should_skip(self) -> None:
        """
        Checks if the first epoch (with index 0) should be skipped to calculate
        the steps per epoch. If the skip is needed, then the internal state
        of the scheduler object will not be changed.
        """
        self._should_skip = False
        if self._update_per_optimizer_step:
            if self._steps_per_epoch is None and self._steps_in_current_epoch > 0:
                self._steps_per_epoch = self._steps_in_current_epoch

            if (
                self._steps_per_epoch is not None
                and self._steps_in_current_epoch > 0
                and self._steps_per_epoch != self._steps_in_current_epoch
            ):
                raise Exception(
                    "Actual steps per epoch and steps per epoch from the scheduler "
                    "parameters are different. Scheduling may be incorrect."
                )

            if self._steps_per_epoch is None:
                self._should_skip = True
                nncf_logger.warning(
                    "Scheduler set to update sparsity level per optimizer step, "
                    "but steps_per_epoch was not set in config. Will only start updating "
                    "sparsity level after measuring the actual steps per epoch as signaled "
                    "by a .epoch_step() call."
                )


@SPARSITY_SCHEDULERS.register("exponential")
class ExponentialSparsityScheduler(SparsityScheduler):
    """
    Sparsity scheduler with an exponential decay schedule.

    This scheduler applies exponential decay to the density level
    to calculate the sparsity level for the `current_epoch`.
    The density level for the `current_epoch` is calculated as

        current_density = 1.0 - current_level
    """

    def __init__(self, controller: SparsityController, params: Dict[str, Any]):
        """
        Initializes a sparsity scheduler with an exponential decay schedule.

        :param controller: Sparsity algorithm controller.
        :param params: Parameters of the scheduler.
        """
        super().__init__(controller, params)

        initial_density = 1.0 - self.initial_level
        target_density = 1.0 - self.target_level
        self.schedule = ExponentialDecaySchedule(initial_density, target_density, self.target_epoch)

    def epoch_step(self, next_epoch: Optional[int] = None) -> None:
        super().epoch_step(next_epoch)
        self._update_sparsity_level()

    def _calculate_sparsity_level(self) -> float:
        current_density: float = self.schedule(self.current_epoch)
        current_level: float = 1.0 - current_density
        return min(current_level, self.target_level)


@SPARSITY_SCHEDULERS.register("adaptive")
class AdaptiveSparsityScheduler(SparsityScheduler):
    """
    Sparsity scheduler with an adaptive schedule.
    """

    def __init__(self, controller: SparsityController, params: Dict[str, Any]):
        """
        Initializes a sparsity scheduler with an adaptive schedule.

        :param controller: Sparsity algorithm controller.
        :param params: Parameters of the scheduler.
        """
        super().__init__(controller, params)
        self.decay_step = params.get("step", 0.05)
        self.eps = params.get("eps", 0.03)
        self.patience = params.get("patience", SPARSITY_SCHEDULER_PATIENCE)
        self.num_bad_epochs = 0
        self._current_level: float = self.initial_level

    @property
    def current_sparsity_level(self) -> float:
        """
        Returns sparsity level for the `current_epoch` or for step
        in the `current_epoch`.

        :return: Current sparsity level.
        """
        return self._current_level

    def epoch_step(self, next_epoch: Optional[int] = None) -> None:
        super().epoch_step(next_epoch)
        self._update_sparsity_level()

    def _calculate_sparsity_level(self) -> float:
        if self._controller.current_sparsity_level >= self._current_level - self.eps:
            self.num_bad_epochs += 1

        current_level = self._current_level
        if self.num_bad_epochs >= self.patience:
            self.num_bad_epochs = 0
            current_level = current_level + self.decay_step

        self._current_level = min(current_level, self.target_level)

        return self._current_level

    def load_state(self, state: Dict[str, Any]) -> None:
        super().load_state(state)
        self.num_bad_epochs = state["num_bad_epochs"]
        self._current_level = state["current_sparsity_level"]

    def get_state(self) -> Dict[str, Any]:
        state = super().get_state()
        state["num_bad_epochs"] = self.num_bad_epochs
        state["current_sparsity_level"] = self._current_level
        return state


@SPARSITY_SCHEDULERS.register("multistep")
class MultiStepSparsityScheduler(SparsityScheduler):
    """
    Sparsity scheduler with a piecewise constant schedule.
    """

    def __init__(self, controller: SparsityController, params: Dict[str, Any]):
        """
        Initializes a sparsity scheduler with a piecewise constant schedule.

        :param controller: Sparsity algorithm controller.
        :param params: Parameters of the scheduler.
        """
        super().__init__(controller, params)
        self.schedule = MultiStepSchedule(
            sorted(params.get("multistep_steps", SPARSITY_MULTISTEP_STEPS)),
            params.get("multistep_sparsity_levels", SPARSITY_MULTISTEP_SPARSITY_LEVELS),
        )
        self.target_level = self.schedule.values[-1]

    @property
    def current_sparsity_level(self) -> float:
        """
        Returns sparsity level for the `current_epoch` or for step
        in the `current_epoch`.

        :return: Current sparsity level.
        """
        if self._current_epoch == -1:
            return self.schedule.values[0]
        return self._calculate_sparsity_level()

    def epoch_step(self, next_epoch: Optional[int] = None) -> None:
        super().epoch_step(next_epoch)
        self._update_sparsity_level()

    def _calculate_sparsity_level(self) -> float:
        return self.schedule(self.current_epoch)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import random
from collections import OrderedDict
from copy import deepcopy
from typing import Any, Callable, Dict, List, Optional, Tuple

import torch
from torch import nn

import nncf
from nncf.common.graph import NNCFNodeName
from nncf.common.graph.layer_attributes import BaseLayerAttributes
from nncf.common.graph.layer_attributes import ConvolutionLayerAttributes
from nncf.common.graph.layer_attributes import GenericWeightedLayerAttributes
from nncf.common.graph.layer_attributes import LinearLayerAttributes
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.graph.transformations.commands import TransformationCommand
from nncf.common.graph.transformations.commands import TransformationPriority
from nncf.common.logging import nncf_logger
from nncf.common.pruning.clusterization import Cluster
from nncf.common.pruning.clusterization import Clusterization
from nncf.common.pruning.mask_propagation import MaskPropagationAlgorithm
from nncf.common.pruning.node_selector import PruningNodeSelector
from nncf.common.pruning.shape_pruning_processor import ShapePruningProcessor
from nncf.common.pruning.structs import PrunedLayerInfoBase
from nncf.common.pruning.utils import get_input_masks
from nncf.common.pruning.utils import get_prunable_layers_in_out_channels
from nncf.common.pruning.utils import is_prunable_depthwise_conv
from nncf.common.scopes import should_consider_scope
from nncf.common.tensor import NNCFTensor
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.base_handler import ELASTICITY_BUILDERS
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.base_handler import ELASTICITY_HANDLERS_MAP
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.base_handler import ELASTICITY_PARAMS
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.base_handler import BaseElasticityParams
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.base_handler import SingleElasticityBuilder
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.base_handler import SingleElasticityHandler
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.elasticity_dim import ElasticityDim
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.filter_reorder import FilterReorderingAlgorithm
from nncf.torch.graph.graph import PTNNCFGraph
from nncf.torch.graph.operator_metatypes import PTModuleBatchNormMetatype
from nncf.torch.graph.operator_metatypes import PTModuleConv2dMetatype
from nncf.torch.graph.operator_metatypes import PTModuleDepthwiseConv2dSubtype
from nncf.torch.graph.operator_metatypes import PTModuleLayerNormMetatype
from nncf.torch.graph.operator_metatypes import PTModuleLinearMetatype
from nncf.torch.graph.transformations.commands import PTInsertionCommand
from nncf.torch.graph.transformations.commands import PTTargetPoint
from nncf.torch.layers import NNCFConv2d
from nncf.torch.layers import NNCFLinear
from nncf.torch.module_operations import UpdateBatchNormParams
from nncf.torch.module_operations import UpdateLayerNormParams
from nncf.torch.module_operations import UpdateNumGroups
from nncf.torch.module_operations import UpdateWeight
from nncf.torch.module_operations import UpdateWeightAndOptionalBias
from nncf.torch.nncf_network import NNCFNetwork
from nncf.torch.pruning.filter_pruning.functions import FILTER_IMPORTANCE_FUNCTIONS
from nncf.torch.pruning.operations import PT_PRUNING_OPERATOR_METATYPES
from nncf.torch.pruning.operations import PTElementwisePruningOp
from nncf.torch.pruning.tensor_processor import PTNNCFPruningTensorProcessor
from nncf.torch.tensor import PTNNCFTensor
from nncf.torch.utils import get_filters_num
from nncf.torch.utils import get_model_device

PruningGroupID = int
WidthType = int
ElasticWidthConfig = Dict[PruningGroupID, WidthType]
WidthList = List[WidthType]
ElasticWidthSearchSpace = Dict[PruningGroupID, WidthList]


def sum_filter(weight_tensor, dim=0):
    """
    Calculates the sum for weight_tensor for the selected dimension.
    """
    weight_tensor = weight_tensor.transpose(0, dim).contiguous()
    return torch.sum(weight_tensor.view(weight_tensor.shape[0], -1), dim=1)


class ElasticWidthOp:
    """
    Base class for introducing elastic width for the operations. On the forward pass it takes parameters of operations
    and trims its input channels (elastic input width) or output channels (elastic output width). This class produces
    2 groups of classes with prefixes ElasticOutputWidth and ElasticInputWidth correspondingly.
    """

    def __init__(self, max_width: int, node_name: str):
        """
        Constructor.

        :param max_width: maximum number of channels in the original layer.
        :param node_name: string representation of operation address. It's used for more informative messages only.
        """
        super().__init__()
        self._active_width = max_width
        self._max_width = max_width
        self._node_name = node_name

    @property
    def max_width(self) -> WidthType:
        """
        :return: maximum number of channels specified on creation of the object.
        """
        return self._max_width

    def get_active_width(self) -> WidthType:
        """
        :return: number of channels to trim on forward call
        """
        return self._active_width

    def set_active_width(self, width: WidthType) -> None:
        """
        Sets current level of elasticity for the operation - number of channels to trim.
        The actual trimming of specified number of channels happens on forward call.
        The value should be less the original width and more than one. Zero number of channels is
        supported through Dynamic Depth feature.

        :param width: number of channels
        """
        if width is None or width > self._max_width or width < 1:
            raise AttributeError(
                "Invalid width={} in scope={}.\nIt should be within the range: [1, {}]".format(
                    width, self._node_name, self._max_width
                )
            )

        self._active_width = width


class EWParamsStateNames:
    MIN_WIDTH = "min_width"
    MAX_NUM_WIDTHS = "max_num_widths"
    WIDTH_STEP = "width_step"
    WIDTH_MULTIPLIERS = "width_multipliers"
    FILTER_IMPORTANCE = "filter_importance"
    EXTERNAL_IMPORTANCE_PATH = "external_importance_path"
    OVERWRITE_GROUPS = "overwrite_groups"
    OVERWRITE_GROUPS_WIDTHS = "overwrite_groups_widths"
    ADD_DYNAMIC_INPUTS = "add_dynamic_inputs"


@ELASTICITY_PARAMS.register(ElasticityDim.WIDTH)
class ElasticWidthParams(BaseElasticityParams):
    _state_names = EWParamsStateNames

    def __init__(
        self,
        min_width: int,
        max_num_widths: int,
        width_step: int,
        width_multipliers: List[float],
        filter_importance: str,
        external_importance_path: Optional[str] = None,
        overwrite_groups: Optional[List[str]] = None,
        overwrite_groups_widths: Optional[List[str]] = None,
        add_dynamic_inputs: Optional[List[str]] = None,
    ):
        """
        Constructor

        :param min_width: Minimal number of output channels that can be activated for each layers with elastic width.
        Default value is 32.
        :param max_num_widths: Restricts total number of different elastic width values for each layer.
        The default value is -1 means that there's no restrictions.
        :param width_step: Defines a step size for a generation of the elastic width search space - the list of all
        possible width values for each layer. The generation starts from the number of output channels in the original
        model and stops when it reaches whether a `min_width` width value or number of generated width values
        equal to `max_num_widths`.
        This parameter is mutually exclusive with `width_multipliers`.
        :param width_multipliers: Defines elastic width search space via a list of multipliers. All possible width
        values are obtained by multiplying the original width value with the values in the given list.
        The obtained values are rounded to the nearest smaller value divisible by alignment constant (e.g. 8).
        This parameter is mutually exclusive with `width_step`.
        :param filter_importance: The type of filter importance metric. Can be one of `L1`, `L2`, `geometric_median`,
        `external`. `L1` by default.
        :param external_importance_path: Path to the custom external weight importance (PyTorch tensor) per node
        that will be used for weight reordering. Valid only when filter_importance is `external`.
        """
        self.min_width = min_width
        self.max_num_widths = max_num_widths
        self.width_step = width_step
        self.width_multipliers = width_multipliers
        assert (
            filter_importance != "external" or external_importance_path is not None
        ), "Missing external weight importance path."
        self.filter_importance = filter_importance
        self.external_importance_path = external_importance_path

        self.overwrite_groups = overwrite_groups
        self.overwrite_groups_widths = overwrite_groups_widths
        self.add_dynamic_inputs = add_dynamic_inputs

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "ElasticWidthParams":
        """
        Creates the object from its config.
        """
        kwargs = {
            cls._state_names.MIN_WIDTH: config.get(cls._state_names.MIN_WIDTH, 32),
            cls._state_names.MAX_NUM_WIDTHS: config.get(cls._state_names.MAX_NUM_WIDTHS, -1),
            cls._state_names.WIDTH_STEP: config.get(cls._state_names.WIDTH_STEP, 32),
            cls._state_names.WIDTH_MULTIPLIERS: config.get(cls._state_names.WIDTH_MULTIPLIERS),
            cls._state_names.FILTER_IMPORTANCE: config.get(cls._state_names.FILTER_IMPORTANCE, "L1"),
            cls._state_names.EXTERNAL_IMPORTANCE_PATH: config.get(cls._state_names.EXTERNAL_IMPORTANCE_PATH, None),
            cls._state_names.OVERWRITE_GROUPS: config.get(cls._state_names.OVERWRITE_GROUPS, None),
            cls._state_names.OVERWRITE_GROUPS_WIDTHS: config.get(cls._state_names.OVERWRITE_GROUPS_WIDTHS, None),
            cls._state_names.ADD_DYNAMIC_INPUTS: config.get(cls._state_names.ADD_DYNAMIC_INPUTS, None),
        }
        return cls(**kwargs)

    @classmethod
    def from_state(cls, state: Dict[str, Any]) -> "ElasticWidthParams":
        """
        Creates the object from its state.

        :param state: Output of `get_state()` method.
        """
        return cls(**state)

    def get_state(self) -> Dict[str, Any]:
        """
        Returns the compression loss state.

        :return: The compression loss state.
        """
        return {
            self._state_names.MIN_WIDTH: self.min_width,
            self._state_names.MAX_NUM_WIDTHS: self.max_num_widths,
            self._state_names.WIDTH_STEP: self.width_step,
            self._state_names.WIDTH_MULTIPLIERS: self.width_multipliers,
            self._state_names.FILTER_IMPORTANCE: self.filter_importance,
            self._state_names.EXTERNAL_IMPORTANCE_PATH: self.external_importance_path,
            self._state_names.OVERWRITE_GROUPS: self.overwrite_groups,
            self._state_names.OVERWRITE_GROUPS_WIDTHS: self.overwrite_groups_widths,
            self._state_names.ADD_DYNAMIC_INPUTS: self.add_dynamic_inputs,
        }

    def __eq__(self, other: "ElasticWidthParams") -> bool:
        return self.__dict__ == other.__dict__

    def __str__(self):
        return (
            f"{self.__class__.__name__}: width_step: {self.width_step} "
            f"min_width: {self.min_width} width_multipliers: {self.width_multipliers} "
            f"max_num_widths: {self.max_num_widths} overwrite_groups: {self.overwrite_groups} "
            f"overwrite_group_widths: {self.overwrite_groups_widths}"
        )


class ElasticOutputWidthOp(ElasticWidthOp):
    """
    Base class for trimming output channels (output width) of the operations.
    """

    def __init__(
        self,
        max_width: int,
        node_name: str,
        params: ElasticWidthParams,
        fixed_width_list: Optional[List[int]] = None,
    ):
        """
        Constructor.

        :param max_width: maximum number of output channels in the original operation.
        :param node_name: string representation of operation address. It's used for more informative messages only.
        :param params: parameters to configure elastic width for the operation.
        """
        super().__init__(max_width=max_width, node_name=node_name)
        if fixed_width_list is None:
            fixed_width_list = []
        if fixed_width_list:
            fixed_width_list.sort(reverse=True)
            if fixed_width_list[0] > max_width:
                raise nncf.InternalError(
                    f"Width list for {node_name} contains invalid values: {fixed_width_list}, {max_width}"
                )
            if fixed_width_list[0] != max_width:
                raise nncf.ValidationError(f"Max width for {node_name} is not aligned with pre-trained model")
            self._width_list = fixed_width_list
        else:
            self._width_list = self._generate_width_list(self._max_width, params)

    @property
    def width_list(self) -> List[int]:
        """
        list of all available widths to select from. Each value corresponds to a single element in the search space of
        operation. The search space of the model is cartesian product of search spaces of operation.
        If all widths starting from 1 to maximum number of channels with step size 1 are available, the search space
        would be prohibitively large to efficiently train and search.

        That's why there are elastic width parameters that constraint number of all available widths.

        :return: list of widths
        """
        return self._width_list

    def set_active_width(self, width: int) -> None:
        """
        Sets current level of elasticity for the operation - number of output channels to trim.
        The actual trimming of specified number of channels happens on forward call.
        The value should be less the original width and more than one. Zero number of channels is
        supported through Dynamic Depth feature.

        :param width: number of output channels
        """
        if width not in self.width_list and width != self.max_width:
            raise ValueError(
                f"Invalid number of output channels to set: {width} in scope={self._node_name}. "
                f"Should be a number in {self.width_list}"
            )
        super().set_active_width(width)

    @staticmethod
    def _generate_width_list(max_width: int, params: ElasticWidthParams) -> List[int]:
        """
        Generates list of available width values.
        There are two mutually exclusive modes: using `width_step` and using `width_multipliers`.
        The mode with `width_step` defines a step size for a generation of the elastic width search space - the list of
        all possible width values for each layer. The generation starts from the number of output channels in the
        original model and stops when it reaches whether a `min_width` width value or number of generated width values
        equal to `max_num_widths`.
        The mode with `width_multipliers` defines elastic width search space via a list of multipliers. All possible
        width values are obtained by multiplying the original width value with the values in the given list.
        The obtained values are rounded to the nearest smaller value divisible by alignment constant (e.g. 8).

        :param max_width: maximum value of width
        :param params: parameters to configure elastic width for the operation.
        :return: list of available width values.
        """
        ALIGNMENT_CONSTANT_FOR_MULTIPLIERS = 8
        width_list = []
        p = params
        if max_width <= p.min_width:
            width_list.append(max_width)
        elif not p.width_multipliers:
            width = max_width
            width_list.append(width)
            width -= p.width_step
            while width >= p.min_width:
                if p.max_num_widths == len(width_list):
                    break
                width_list.append(width)
                width -= p.width_step
        else:
            p.width_multipliers.sort(reverse=True)
            if p.width_multipliers[0] < 1:
                width_list.append(max_width)
            for multiplier in p.width_multipliers:
                if p.max_num_widths == len(width_list):
                    break
                if 0 >= multiplier > 1:
                    raise nncf.InternalError(f"Wrong value for multiplier: {multiplier}")
                w = int(max_width * multiplier)
                w = w - (w % ALIGNMENT_CONSTANT_FOR_MULTIPLIERS)
                w = max(w, p.min_width)
                if w in width_list:
                    continue
                width_list.append(w)
        return width_list


class ElasticWidthInfo(PrunedLayerInfoBase):
    """
    List of attributes describing operation with elastic width
    """

    def __init__(
        self,
        node_name: NNCFNodeName,
        module: nn.Module,
        elastic_op: ElasticOutputWidthOp,
        node_id: int,
        is_depthwise: bool,
    ):
        super().__init__(node_name, node_id, is_depthwise)
        self.module = module
        self.elastic_op = elastic_op

    def __str__(self):
        return (
            f"{self.__class__.__name__}: node_name: {self.node_name} module: {self.module} "
            f"elastic_op: {self.elastic_op} node_id: {self.nncf_node_id} is_depthwise: {self.is_depthwise}"
        )


class ElasticInputWidthLinearOp(ElasticWidthOp, nn.Module):
    """
    Introduces elastic input width for linear layer.
    """

    def forward(self, weight: torch.Tensor) -> torch.Tensor:
        """
        Trims weight according to active number of input channels

        :param weight: weight tensor to be trimmed
        :return: trimmed weight
        """
        return weight[:, : self._active_width]


class ElasticInputWidthConvOp(ElasticWidthOp, nn.Module):
    """
    Introduces elastic input width for 2D convolution.
    """

    def forward(self, weight: torch.Tensor) -> torch.Tensor:
        """
        Trims weight according to active number of input channels

        :param weight: weight tensor to be trimmed
        :return: trimmed weight
        """
        return weight[:, : self._active_width, :, :]


class ElasticInputWidthDWConvOp(ElasticWidthOp, nn.Module):
    """
    Introduces elastic input width for depthwise convolution.
    """

    def forward(self, _) -> int:
        """
        :return: number of input channels to be trimmed. In case of depthwise convolution no need to trim weights, just
        need to change number of group accordingly.
        """
        return self._active_width


class ElasticInputWidthBatchNormOp(ElasticWidthOp, nn.Module):
    """
    Introduces elastic input width for batchnorm layer.
    """

    SET_RUNNING_STATISTICS = False

    def forward(self, **bn_params: torch.Tensor) -> List[torch.Tensor]:
        """
        Trims batchnorm parameters according to active number of input channels.

        :param bn_params: map of name and tensor to be trimmed
        :return: trimmed batchnorm parameters
        """
        return [param[: self._active_width] for param in bn_params.values()]


class ElasticInputWidthLayerNormOp(ElasticWidthOp, nn.Module):
    """
    Introduces elastic input width for layernorm layer.
    """

    def forward(self, weight: torch.Tensor, bias: torch.Tensor, normalized_shape: torch.Tensor) -> List[torch.Tensor]:
        """
        Trims layernorm parameters according to active number of input channels.
        :param weight: weight tensor to be trimmed
        :param bias: bias tensor to be trimmed
        :param normalized_shape: normalized_shape to be trimmed
        :return: list of trimmed layernorm parameters
        """
        assert len(normalized_shape) == 1, "Currently only 1-dimensional shape is supported."
        return [
            weight[: self._active_width],
            bias[: self._active_width],
            (self._active_width,),
        ]


class ElasticOutputWidthConv2DOp(ElasticOutputWidthOp, nn.Module):
    """
    Introduces elastic output width for 2D convolution.
    """

    def forward(self, weight: torch.Tensor, bias: Optional[torch.Tensor]) -> List[torch.Tensor]:
        """
        Trims convolution parameters according to active number of output channels.

        :param weight: weight tensor to be trimmed
        :param bias: bias tensor to be trimmed
        :return: list of trimmed convolution parameters
        """
        nncf_logger.debug(f"Conv2d with active width={self._active_width} in scope={self._node_name}")
        num_out_channels = self._active_width
        new_bias = None if bias is None else bias[:num_out_channels]
        new_weights = weight[:num_out_channels, :, :, :]
        return [new_weights, new_bias]


class ElasticOutputWidthLinearOp(ElasticOutputWidthOp, nn.Module):
    """
    Introduces elastic output width for linear layer.
    """

    def forward(self, weight: torch.Tensor, bias: Optional[torch.Tensor]) -> List[torch.Tensor]:
        """
        Trims linear layer's parameters according to active number of output channels.

        :param weight: weight tensor to be trimmed
        :param bias: bias tensor to be trimmed
        :return: list of trimmed linear parameters
        """
        new_bias = None if bias is None else bias[: self._active_width]
        return [weight[: self._active_width, :], new_bias]


class EWHandlerStateNames:
    WIDTH_NUM_PARAMS_INDICATOR = "width_num_params_indicator"


class ElasticWidthHandler(SingleElasticityHandler):
    """
    An interface for handling elastic width dimension in the network, i.e. define number of channels in the layers.
    """

    _width_state_names = EWHandlerStateNames

    def __init__(
        self,
        target_model: NNCFNetwork,
        filter_importance_fn: Callable[[torch.Tensor, int], torch.Tensor],
        external_importance_path: Optional[str],
        weights_normalizer_fn: Optional[Callable[[torch.Tensor], torch.Tensor]],
        node_name_vs_dynamic_input_width_op_map: Dict[NNCFNodeName, ElasticWidthOp],
        pruned_module_groups_info: Clusterization[ElasticWidthInfo](id_fn=lambda x: x.node_name),
        transformation_commands: List[TransformationCommand],
        add_dynamic_inputs: Optional[List[str]] = None,
    ):
        """
        Constructor

        :param target_model: a target NNCFNetwork for adding elasticity.
        :param filter_importance_fn: a callable that implements calculation of the importance of filters along a given
        dimension for a given weight tensor.
        :param weights_normalizer_fn: a callable that implements normalization of weight tensor
        :param node_name_vs_dynamic_input_width_op_map: map of node name to elastic width operation associated with it
        :param pruned_module_groups_info: structure that specifies information about groups of modules being pruned
        using elastic width operations.
        :param transformation_commands: list of transformation commands for introducing elastic width operations.
        """
        super().__init__()
        self._target_model = target_model
        self._node_name_vs_dynamic_input_width_op_map = node_name_vs_dynamic_input_width_op_map
        self._pruned_module_groups_info = pruned_module_groups_info
        self._transformation_commands = transformation_commands
        self._filter_importance_fn = filter_importance_fn
        self._external_importance = None
        if external_importance_path is not None:
            self._external_importance = torch.load(external_importance_path)
            nncf_logger.debug("Loaded custom external weight importance.")
        self._weights_normalizer_fn = weights_normalizer_fn
        self._add_dynamic_inputs = add_dynamic_inputs

        graph = self._target_model.nncf.get_original_graph()
        prunable_types = [NNCFConv2d.op_func_name, NNCFLinear.op_func_name]
        self._shape_pruning_processor = ShapePruningProcessor(
            prunable_types=prunable_types,
            pruning_operations_metatype=PT_PRUNING_OPERATOR_METATYPES,
        )
        self._next_nodes = self._shape_pruning_processor.get_next_nodes(graph, pruned_module_groups_info)
        # Need a copy because it will be used for adding `output_mask`/`input_masks` to nodes that are relevant to
        # Elastic Width only and therefore it should be isolated to not intercept with other algorithms.
        self._propagation_graph = deepcopy(graph)

        self._width_num_params_indicator = -1

    @property
    def width_num_params_indicator(self):
        return self._width_num_params_indicator

    @width_num_params_indicator.setter
    def width_num_params_indicator(self, width_num_params_indicator):
        if width_num_params_indicator == 0 or width_num_params_indicator < -1:
            raise nncf.InternalError(f"Invalid width indicator: {width_num_params_indicator}")
        self._width_num_params_indicator = width_num_params_indicator

    @property
    def propagation_graph(self) -> PTNNCFGraph:
        """
        :return: nncf graph that is used for propagating pruning and reordering masks
        """
        return self._propagation_graph

    def load_state(self, state: Dict[str, Any]) -> None:
        """
        Initializes object from the state.
        :param state: Output of `get_state()` method.
        """
        super().load_state(state)
        self.width_num_params_indicator = state[self._width_state_names.WIDTH_NUM_PARAMS_INDICATOR]

    def get_state(self) -> Dict[str, Any]:
        """
        Returns a dictionary with Python data structures (dict, list, tuple, str, int, float, True, False, None) that
        represents state of the object.
        :return: state of the object
        """
        state = super().get_state()
        state[self._width_state_names.WIDTH_NUM_PARAMS_INDICATOR] = self.width_num_params_indicator
        return state

    def get_transformation_commands(self) -> List[TransformationCommand]:
        """
        :return: transformation commands for introducing the elasticity to NNCFNetwork
        """
        return self._transformation_commands

    def get_search_space(self) -> ElasticWidthSearchSpace:
        """
        :return: search space that is produced by iterating over all elastic parameters
        """
        if self._width_num_params_indicator == -1:
            return self._collect_ops_data_by_selection_rule(lambda op: op.width_list)
        return self._collect_ops_data_by_selection_rule(
            lambda op: op.width_list[: min(self._width_num_params_indicator, len(op.width_list))]
        )

    def get_active_config(self) -> ElasticWidthConfig:
        """
        Forms an elasticity configuration that describes currently activated Subnet

        :return: map of pruning group id to width value
        """
        return self._collect_ops_data_by_selection_rule(lambda op: op.get_active_width())

    def get_random_config(self) -> ElasticWidthConfig:
        """
        Forms an elasticity configuration that describes a Subnet with randomly chosen width for each elastic layer

        :return: map of pruning group id to width value
        """
        return self._collect_ops_data_by_selection_rule(
            lambda op: op.width_list[random.randrange(0, self._get_width_list_len(op))]  # nosec
        )

    def get_minimum_config(self) -> ElasticWidthConfig:
        """
        Forms an elasticity configuration that describes a Subnet with minimum width in each elastic layer

        :return: map of pruning group id to width value
        """
        return self._collect_ops_data_by_selection_rule(lambda op: min(self._get_width_list(op)))

    def get_maximum_config(self) -> ElasticWidthConfig:
        """
        Forms an elasticity configuration that describes a Subnet with maximum width in each elastic layer

        :return: map of pruning group id to width value
        """
        return self._collect_ops_data_by_selection_rule(lambda op: max(self._get_width_list(op)))

    def activate_supernet(self) -> None:
        """
        Activates the Supernet - the original network to which elasticity was applied.
        """
        supernet_config = self._collect_ops_data_by_selection_rule(lambda op: op.max_width)
        self.activate_subnet_for_config(supernet_config)

    def activate_subnet_for_config(self, config: ElasticWidthConfig) -> None:
        """
        Activates a Subnet that corresponds to the given elasticity configuration

        :param config: map of pruning group id to width value
        """
        for node in self._propagation_graph.get_all_nodes():
            node.attributes.pop("output_mask", None)

        names_of_processed_nodes = set()
        for cluster_id, width in config.items():
            cluster = self._pruned_module_groups_info.get_cluster_by_id(cluster_id)
            for elastic_width_info in cluster.elements:
                node_id = elastic_width_info.nncf_node_id
                node = self._propagation_graph.get_node_by_id(node_id)
                max_width = elastic_width_info.elastic_op.max_width
                device = get_model_device(self._target_model)
                mask = self._width_to_mask(width, max_width, device)
                node.attributes["output_mask"] = mask
                elastic_width_info.elastic_op.set_active_width(width)
                names_of_processed_nodes.add(node_id)

        algo = MaskPropagationAlgorithm(
            self._propagation_graph,
            PT_PRUNING_OPERATOR_METATYPES,
            PTNNCFPruningTensorProcessor,
        )
        algo.mask_propagation()

        for (
            node_name,
            dynamic_input_width_op,
        ) in self._node_name_vs_dynamic_input_width_op_map.items():
            node = self._propagation_graph.get_node_by_name(node_name)
            input_masks = get_input_masks(node, self._propagation_graph)
            was_set = False
            if input_masks:
                input_mask = input_masks[0]
                input_width = self.mask_to_width(input_mask)
                if input_width:
                    dynamic_input_width_op.set_active_width(input_width)
                    was_set = True

            if not was_set and node_name not in names_of_processed_nodes:
                nncf_logger.debug(f"input width was not set in scope={node.node_name}")

            if self._add_dynamic_inputs and node_name in self._add_dynamic_inputs and not was_set:
                nncf_logger.debug(f"setting input width by user's request for scope={node_name}")
                nodes_to_check = [node]
                while any(elem is None for elem in input_masks):
                    previous_nodes = []
                    for node in nodes_to_check:
                        previous_nodes.append(self._propagation_graph.get_previous_nodes(node))
                    nodes_to_check.clear()
                    previous_nodes = [item for nodes in previous_nodes for item in nodes]
                    if not previous_nodes:
                        break
                    for previous in previous_nodes:
                        if "output_mask" in previous.attributes:
                            if previous.attributes["output_mask"] is not None:
                                input_masks.append(previous.attributes["output_mask"])
                                input_masks = [i for i in input_masks if i]
                            else:
                                nodes_to_check.append(previous)
                        else:
                            nodes_to_check.append(previous)
                if input_masks:
                    input_mask = input_masks[0]
                    input_width = self.mask_to_width(input_mask)
                    if input_width:
                        dynamic_input_width_op.set_active_width(input_width)
                        was_set = True
                if was_set:
                    nncf_logger.debug(f"Success setting up user's request for dynamic input at scope={node_name}")

    def get_active_in_out_width_values(
        self,
    ) -> Tuple[Dict[NNCFNodeName, int], Dict[NNCFNodeName, int]]:
        """
        Collects the active number of input and output channels (width) for each elastic layer in the graph.

        :return Dictionary with the number of input channels to convolution and linear layers:
            {node_name: input_channels_num}
            Dictionary with the number of output channels from convolution and linear layers:
            {node_name: output_channels_num}
        """
        graph = self._target_model.nncf.get_graph()
        in_channels, out_channels = get_prunable_layers_in_out_channels(graph)

        for group in self._pruned_module_groups_info.get_all_clusters():
            assert all(
                out_channels[group.elements[0].node_name] == out_channels[node.node_name] for node in group.elements
            )
            first_elastic_width_info: ElasticWidthInfo = group.elements[0]
            first_elastic_op: ElasticOutputWidthOp = first_elastic_width_info.elastic_op
            new_out_channels_num = first_elastic_op.get_active_width()
            num_of_pruned_elems = first_elastic_op.max_width - new_out_channels_num
            self._shape_pruning_processor.prune_cluster_shapes(
                group, num_of_pruned_elems, self._next_nodes, in_channels, out_channels
            )

        return in_channels, out_channels

    def resolve_conflicts_with_other_elasticities(
        self, config: ElasticWidthConfig, elasticity_handlers: ELASTICITY_HANDLERS_MAP
    ) -> ElasticWidthConfig:
        """
        Resolves a conflict between the given elasticity config and active elasticity configs of the given handlers.
        For example, elastic width configuration may contradict to elastic depth one. When we activate some
        configuration in the Elastic Width Handler, i.e. define number of output channels for some layers, we
        change output shapes of the layers. Consequently, it affects the blocks that can be skipped by Elastic Depth
        Handler, because input and output shapes may not be identical now.

        :param config: elasticity configuration
        :param elasticity_handlers: map of elasticity dimension to elasticity handler
        :return: elasticity configuration without conflicts with other active configs of other elasticity handlers
        """
        return config

    def get_group_id_by_node_name(self, node_name: NNCFNodeName) -> Optional[int]:
        """
        Provides a pruning group number that corresponds to a layer with the given name in graph. It's intended to be
        used for logging and visualization only.

        :param node_name: node name
        :return: group id
        """
        group_id = None
        for cluster in self._pruned_module_groups_info.get_all_clusters():
            for element in cluster.elements:
                if node_name == element.node_name:
                    group_id = cluster.id
        return group_id

    def get_external_importance(self, node_name: NNCFNodeName):
        """
        Return custom weight importance for the current node from external source.

        :param node_name: node name
        :return: importance tensor
        """
        assert should_consider_scope(
            node_name, ignored_scopes=None, target_scopes=self._external_importance.keys()
        ), f"Cannot match {node_name} in external weight importance data structure"
        return self._external_importance[node_name]

    def reorganize_weights(self) -> None:
        """
        Reorder output filters in descending order of their importance.
        """
        for node in self._propagation_graph.get_all_nodes():
            node.attributes.pop("output_mask", None)

        # 1. Calculate filter importance for all groups of prunable layers
        for group in self._pruned_module_groups_info.get_all_clusters():
            filters_num = torch.tensor([get_filters_num(minfo.module) for minfo in group.elements])
            assert torch.all(filters_num == filters_num[0])
            device = group.elements[0].module.weight.device

            cumulative_filters_importance = torch.zeros(filters_num[0]).to(device)
            # 1.1 Calculate cumulative importance for all filters in this group
            for minfo in group.elements:
                weight = minfo.module.weight
                if self._weights_normalizer_fn:
                    weight = self._weights_normalizer_fn(minfo.module.weight)
                if self._external_importance is not None:
                    weight = self.get_external_importance(minfo.node_name).to(device)
                filters_importance = self._filter_importance_fn(weight, minfo.module.target_weight_dim_for_compression)
                cumulative_filters_importance += filters_importance

            _, reorder_indexes = torch.sort(cumulative_filters_importance, dim=0, descending=True)
            device = get_model_device(self._target_model)
            reorder_indexes.to(device)
            # 1.2 Setup reorder indexes as output mask to reorganize filters
            for minfo in group.elements:
                node = self._propagation_graph.get_node_by_id(minfo.nncf_node_id)
                node.attributes["output_mask"] = PTNNCFTensor(reorder_indexes)

        # 2. Propagating masks across the graph
        reorder_algo = FilterReorderingAlgorithm(
            self._target_model,
            self._propagation_graph,
            PT_PRUNING_OPERATOR_METATYPES,
            PTNNCFPruningTensorProcessor,
        )
        reorder_algo.reorder_filters()

    def find_pairs_of_nodes_with_different_width(self, pairs_of_nodes: List[Tuple[str, str]]) -> List[int]:
        """
        Find pairs of nodes that have different output shapes. It's need to resolve conflict between elastic width and
        elastic depth. The conflicting situation happens when layers are trimmed and skipped independently.
        There might be a situation when channels are trimmed for the layers in the beginning of skipped block, and
        not trimmed in the end, thus violating the necessary condition for skipping a block - output shapes on
        block boundaries should be the same.

        :param pairs_of_nodes: list of pairs of node names
        :return: index of nodes pair in the given list that have different output shapes.
        """
        pair_indexes = []
        for idx, (start_node_name, end_node_name) in enumerate(pairs_of_nodes):
            start_node = self._propagation_graph.get_node_by_name(start_node_name)
            start_mask = start_node.attributes["output_mask"]
            end_node = self._propagation_graph.get_node_by_name(end_node_name)
            end_mask = end_node.attributes["output_mask"]

            all_start_output_shapes = self._propagation_graph.get_output_shapes_for_node(start_node_name)
            start_output_shape = list(OrderedDict.fromkeys(all_start_output_shapes))
            all_end_output_shape = self._propagation_graph.get_output_shapes_for_node(end_node_name)
            end_output_shape = list(OrderedDict.fromkeys(all_end_output_shape))

            start_width = ElasticWidthHandler.mask_to_width(start_mask)
            end_width = ElasticWidthHandler.mask_to_width(end_mask)

            if start_width is None and end_width is None and start_output_shape != end_output_shape:
                reason = f"it has a different shapes on boundaries: {start_output_shape} != {end_output_shape}"
            elif start_width is not None and end_width is not None and start_width != end_width:
                reason = f"it has a different width on boundaries: {start_width} != {end_width}"
            elif (start_width is None or end_width is None) and not (start_width is None and end_width is None):
                reason = (
                    f"it has empty width on one of the boundaries. "
                    f"Width: {start_width} vs {end_width}. Shapes: {start_output_shape} vs {end_output_shape}"
                )
            else:
                continue
            nncf_logger.debug(
                f"The block [\n\t{start_node_name},\n\t{end_node_name}\n]\n can`t be skipped, because {reason}"
            )
            pair_indexes.append(idx)
        return pair_indexes

    def _get_width_list_len(self, op: ElasticOutputWidthOp) -> int:
        N = len(op.width_list)
        if 0 < self._width_num_params_indicator < N:
            return self._width_num_params_indicator
        return N

    def _get_width_list(self, op: ElasticOutputWidthOp) -> List[int]:
        width_list_len = self._get_width_list_len(op)
        return op.width_list[:width_list_len]

    def _collect_ops_data_by_selection_rule(self, selection_rule: Callable) -> Dict[PruningGroupID, Any]:
        elastic_width_config = {}
        for cluster in self._pruned_module_groups_info.get_all_clusters():
            all_max_out_channels = {el.elastic_op.max_width for el in cluster.elements}
            if len(all_max_out_channels) != 1:
                raise nncf.InternalError("Invalid grouping of layers with different number of output channels")

            first_elastic_width_info = next(iter(cluster.elements))
            op = first_elastic_width_info.elastic_op
            selected_width = selection_rule(op)
            elastic_width_config[cluster.id] = selected_width
            nncf_logger.debug(f"Select width={cluster.id} for group #{selected_width}")
        return elastic_width_config

    @staticmethod
    def mask_to_width(mask: NNCFTensor) -> Optional[int]:
        """
        Decodes mask to a single integer. We assume that mask was constructed in a way that first N values are equal
        to 1, and the rest values are 0. The N encodes width value.

        :param mask: tensor with 1 and 0 values.
        :return: width value
        """
        result = None
        if mask is not None:
            actual_mask = mask.tensor
            mask_len = sum(actual_mask.size())
            width = int(sum(actual_mask))
            device = actual_mask.device
            ref_mask = ElasticWidthHandler._width_to_mask(width, mask_len, device).tensor
            assert torch.equal(
                ref_mask, actual_mask
            ), f"Invalid mask {actual_mask}: the first {width} values must be ones, the rest - zeros."
            result = width
        return result

    @staticmethod
    def _width_to_mask(active_width: int, max_width: int, device: torch.device) -> PTNNCFTensor:
        """
        Encodes width to tensor filled by 1 and 0. We intentionally construct mask in a way that first N values are
        equal to 1, and the rest values are 0. The N encodes width value. The mask tensors allows to fully reuse mask
        propagation algorithm without any changes, in contrast of implementing a separate width propagation algorithm.

        :param active_width: width value to encode in the mask.
        :param max_width: maximum width value.
        :param device: device that should have a mask tensor.
        :return: encoded mask tensor
        """
        mask = torch.ones(max_width).to(device)
        mask[active_width:].fill_(0)
        return PTNNCFTensor(mask)


class EWBuilderStateNames:
    GROUPED_NODE_NAMES_TO_PRUNE = "grouped_node_names_to_prune"
    OVERWRITE_GROUP_WIDTHS = "overwrite_groups_widths"
    ADD_DYNAMIC_INPUTS = "add_dynamic_inputs"


@ELASTICITY_BUILDERS.register(ElasticityDim.WIDTH)
class ElasticWidthBuilder(SingleElasticityBuilder):
    """
    Determines which modifications should be made to the original FP32 model in order to introduce elastic width
    to the model.
    """

    _state_names = EWBuilderStateNames

    def __init__(
        self,
        params: ElasticWidthParams,
        ignored_scopes: Optional[List[str]] = None,
        target_scopes: Optional[List[str]] = None,
    ):
        super().__init__(params, ignored_scopes, target_scopes)
        self._weights_normalizer = None
        self._overwriting_pruning_groups = params.overwrite_groups is not None
        self._grouped_node_names_to_prune: List[List[NNCFNodeName]] = (
            params.overwrite_groups if params.overwrite_groups is not None else []
        )
        self._overwrite_groups_widths = params.overwrite_groups_widths
        self._add_dynamic_inputs = params.add_dynamic_inputs
        self._params = params

    def build(self, target_model: NNCFNetwork) -> ElasticWidthHandler:
        """
        Creates modifications to the given NNCFNetwork for introducing elastic width and creates a handler object that
        can manipulate this elasticity.

        :param target_model: a target NNCFNetwork for adding modifications
        :return: a handler object that can manipulate the elastic width.
        """
        filter_importance_str = self._params.filter_importance
        filter_importance = FILTER_IMPORTANCE_FUNCTIONS.get(filter_importance_str, sum_filter)
        external_importance_path = self._params.external_importance_path

        graph = target_model.nncf.get_original_graph()
        device = next(target_model.parameters()).device

        if not self._grouped_node_names_to_prune:
            prunable_types = [NNCFConv2d, NNCFLinear]
            prune_operations_types = [pt.op_func_name for pt in prunable_types]
            types_of_grouping_ops = PTElementwisePruningOp.get_all_op_aliases()
            pruning_node_selector = PruningNodeSelector(
                PT_PRUNING_OPERATOR_METATYPES,
                prune_operations_types,
                types_of_grouping_ops,
                ignored_scopes=self._ignored_scopes,
                target_scopes=self._target_scopes,
                prune_first=True,
                prune_downsample_convs=True,
            )
            groups_of_nodes_to_prune = pruning_node_selector.create_pruning_groups(graph)
            for group in groups_of_nodes_to_prune.get_all_clusters():
                grouped_node_names = [node.node_name for node in group.elements]
                self._grouped_node_names_to_prune.append(grouped_node_names)

        transformation_commands = []
        pruned_module_groups_info = Clusterization[ElasticWidthInfo](id_fn=lambda x: x.node_name)
        node_name_vs_dynamic_input_width_op_map = OrderedDict()

        metatype_vs_elastic_op_creator = {
            PTModuleConv2dMetatype: self._create_elastic_conv_width_op,
            PTModuleDepthwiseConv2dSubtype: self._create_elastic_conv_width_op,
            PTModuleLinearMetatype: self._create_elastic_linear_width_op,
        }

        for i, grouped_node_names in enumerate(self._grouped_node_names_to_prune):
            group_minfos = []
            list_of_node_ids = []
            for node_name in grouped_node_names:
                node = graph.get_node_by_name(node_name)
                metatype = node.metatype
                list_of_node_ids.append(node.node_id)
                layer_attrs = node.layer_attributes
                if metatype not in metatype_vs_elastic_op_creator:
                    raise nncf.InternalError(f"Elastic width is not supported for {metatype}")
                elastic_op_creator = metatype_vs_elastic_op_creator[metatype]

                elastic_width_operation = elastic_op_creator(
                    layer_attrs,
                    node_name,
                    self._params,
                    self._overwrite_groups_widths[i] if self._overwriting_pruning_groups else [],
                )
                elastic_width_operation.to(device)
                update_conv_params_op = UpdateWeightAndOptionalBias(elastic_width_operation)
                transformation_commands.append(
                    PTInsertionCommand(
                        PTTargetPoint(TargetType.PRE_LAYER_OPERATION, target_node_name=node_name),
                        update_conv_params_op,
                        TransformationPriority.PRUNING_PRIORITY,
                    )
                )
                pruned_module = target_model.nncf.get_containing_module(node_name)
                assert isinstance(
                    pruned_module, (nn.Conv2d, nn.Linear)
                ), "currently prune only 2D Convolutions and Linear layers"

                group_minfos.append(
                    ElasticWidthInfo(
                        node_name=node_name,
                        module=pruned_module,
                        elastic_op=elastic_width_operation,
                        node_id=node.node_id,
                        is_depthwise=is_prunable_depthwise_conv(node),
                    )
                )

            cluster = Cluster[ElasticWidthInfo](i, group_minfos, list_of_node_ids)
            pruned_module_groups_info.add_cluster(cluster)

        metatype_vs_dynamic_input_op_creator = {
            PTModuleConv2dMetatype: self._create_dynamic_conv_input_op,
            PTModuleDepthwiseConv2dSubtype: self._create_dynamic_dw_conv_input_op,
            PTModuleBatchNormMetatype: self._create_dynamic_bn_input_op,
            PTModuleLayerNormMetatype: self._create_dynamic_ln_input_op,
            PTModuleLinearMetatype: self._create_dynamic_linear_input_op,
        }
        for metatype, op_creator in metatype_vs_dynamic_input_op_creator.items():
            nodes = graph.get_nodes_by_metatypes([metatype])
            for node in nodes:
                node_name = node.node_name
                nncf_logger.debug(f"Adding Dynamic Input Op for {metatype.name} in scope: {node_name}")
                layer_attrs = node.layer_attributes
                update_module_params = op_creator(layer_attrs, node_name).to(device)
                node_name_vs_dynamic_input_width_op_map[node_name] = update_module_params.op
                transformation_commands.append(
                    PTInsertionCommand(
                        PTTargetPoint(TargetType.PRE_LAYER_OPERATION, target_node_name=node_name),
                        update_module_params,
                        priority=TransformationPriority.DEFAULT_PRIORITY,
                    )
                )

        return ElasticWidthHandler(
            target_model,
            filter_importance,
            external_importance_path,
            self._weights_normalizer,
            node_name_vs_dynamic_input_width_op_map,
            pruned_module_groups_info,
            transformation_commands,
            self._add_dynamic_inputs,
        )

    def load_state(self, state: Dict[str, Any]) -> None:
        """
        Initializes object from the state.

        :param state: Output of `get_state()` method.
        """
        params_from_state = state[SingleElasticityBuilder._state_names.ELASTICITY_PARAMS]
        params = ElasticWidthParams.from_state(params_from_state)
        if self._params and self._params != params:
            nncf_logger.warning(
                "Different elasticity parameters were provided in two places: on init and on loading "
                "state. The one from state is taken by ignoring the ones from init."
            )
        self._params = params
        self._grouped_node_names_to_prune = state[self._state_names.GROUPED_NODE_NAMES_TO_PRUNE]

        if params_from_state.get(self._state_names.OVERWRITE_GROUP_WIDTHS, None) is not None:
            self._overwrite_groups_widths = params_from_state[self._state_names.OVERWRITE_GROUP_WIDTHS]
            self._overwriting_pruning_groups = True
            if len(self._grouped_node_names_to_prune) != len(self._overwrite_groups_widths):
                raise nncf.InternalError("Mismatch between number of groups for pruning and their corresponding widths")
        if params_from_state.get(self._state_names.ADD_DYNAMIC_INPUTS, None) is not None:
            self._add_dynamic_inputs = params_from_state[self._state_names.ADD_DYNAMIC_INPUTS]

    def get_state(self) -> Dict[str, Any]:
        """
        Returns a dictionary with Python data structures (dict, list, tuple, str, int, float, True, False, None) that
        represents state of the object.

        :return: state of the object
        """
        return {
            SingleElasticityBuilder._state_names.ELASTICITY_PARAMS: self._params.get_state(),
            self._state_names.GROUPED_NODE_NAMES_TO_PRUNE: self._grouped_node_names_to_prune,
        }

    @staticmethod
    def _create_elastic_conv_width_op(
        conv_layer_attrs: BaseLayerAttributes,
        node_name: str,
        params: ElasticWidthParams,
        fixed_width_list: Optional[List[int]] = None,
    ) -> ElasticOutputWidthConv2DOp:
        assert isinstance(conv_layer_attrs, ConvolutionLayerAttributes)
        nncf_logger.debug(f"Adding Dynamic Conv2D Layer in scope: {str(node_name)}")
        if fixed_width_list is None:
            fixed_width_list = []
        return ElasticOutputWidthConv2DOp(
            conv_layer_attrs.out_channels,
            node_name,
            params,
            fixed_width_list=fixed_width_list,
        )

    @staticmethod
    def _create_elastic_linear_width_op(
        linear_layer_attrs: BaseLayerAttributes,
        node_name: str,
        params: ElasticWidthParams,
        fixed_width_list: Optional[List[int]] = None,
    ) -> ElasticOutputWidthLinearOp:
        assert isinstance(linear_layer_attrs, LinearLayerAttributes)
        if fixed_width_list is None:
            fixed_width_list = []
        nncf_logger.debug(f"Adding Dynamic Linear Layer in scope: {str(node_name)}")
        return ElasticOutputWidthLinearOp(
            linear_layer_attrs.out_features,
            node_name,
            params,
            fixed_width_list=fixed_width_list,
        )

    @staticmethod
    def _create_dynamic_conv_input_op(conv_layer_attrs: BaseLayerAttributes, node_name: str) -> UpdateWeight:
        assert isinstance(conv_layer_attrs, ConvolutionLayerAttributes)
        dynamic_conv_input_op = ElasticInputWidthConvOp(max_width=conv_layer_attrs.in_channels, node_name=node_name)
        return UpdateWeight(dynamic_conv_input_op)

    @staticmethod
    def _create_dynamic_dw_conv_input_op(conv_layer_attrs: BaseLayerAttributes, node_name: str) -> UpdateNumGroups:
        assert isinstance(conv_layer_attrs, ConvolutionLayerAttributes)
        dynamic_dw_conv_input_op = ElasticInputWidthDWConvOp(max_width=conv_layer_attrs.groups, node_name=node_name)
        return UpdateNumGroups(dynamic_dw_conv_input_op)

    @staticmethod
    def _create_dynamic_bn_input_op(generic_layer_attrs: BaseLayerAttributes, node_name: str) -> UpdateBatchNormParams:
        assert isinstance(generic_layer_attrs, GenericWeightedLayerAttributes)
        dynamic_bn_input_op = ElasticInputWidthBatchNormOp(
            max_width=generic_layer_attrs.get_num_filters(), node_name=node_name
        )
        return UpdateBatchNormParams(dynamic_bn_input_op)

    @staticmethod
    def _create_dynamic_ln_input_op(generic_layer_attrs: BaseLayerAttributes, node_name: str) -> UpdateLayerNormParams:
        assert isinstance(generic_layer_attrs, GenericWeightedLayerAttributes)
        dynamic_ln_input_op = ElasticInputWidthLayerNormOp(
            max_width=generic_layer_attrs.get_num_filters(), node_name=node_name
        )
        return UpdateLayerNormParams(dynamic_ln_input_op)

    @staticmethod
    def _create_dynamic_linear_input_op(linear_layer_attrs: BaseLayerAttributes, node_name: str) -> UpdateWeight:
        assert isinstance(linear_layer_attrs, LinearLayerAttributes)
        dynamic_linear_input_op = ElasticInputWidthLinearOp(
            max_width=linear_layer_attrs.in_features, node_name=node_name
        )
        return UpdateWeight(dynamic_linear_input_op)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import tensorflow as tf
from tensorflow.python.framework.convert_to_constants import convert_variables_to_constants_v2

from nncf.common.exporter import Exporter
from nncf.telemetry import tracked_function
from nncf.telemetry.events import NNCF_TF_CATEGORY


class TFExportFormat:
    SAVED_MODEL = "tf"
    KERAS_H5 = "h5"
    FROZEN_GRAPH = "frozen_graph"


# TODO(andrey-churkin): Add support for `input_names` and `output_names`
class TFExporter(Exporter):
    """
    This class provides export of the compressed model to the Frozen Graph,
    TensorFlow SavedModel, or Keras H5 formats.
    """

    @tracked_function(NNCF_TF_CATEGORY, ["save_format"])
    def export_model(self, save_path: str, save_format: str = TFExportFormat.FROZEN_GRAPH) -> None:
        """
        Exports the compressed model to the specified format.

        :param save_path: The path where the model will be saved.
        :param save_format: Saving format.
            One of the following:
                - `tf` for export to the Tensorflow SavedModel format.
                - `h5` for export to the Keras H5 format.
                - `frozen_graph` for export to the Frozen Graph format.
            The Frozen Graph format will be used if `save_format` is not specified.
        """

        format_to_export_fn = {
            TFExportFormat.SAVED_MODEL: self._export_to_saved_model,
            TFExportFormat.KERAS_H5: self._export_to_h5,
            TFExportFormat.FROZEN_GRAPH: self._export_to_frozen_graph,
        }

        export_fn = format_to_export_fn.get(save_format)

        if export_fn is None:
            available_formats = list(format_to_export_fn.keys())
            raise ValueError(f"Unsupported saving format: '{save_format}'. Available formats: {available_formats}")

        export_fn(save_path)

    def _export_to_saved_model(self, save_path: str) -> None:
        """
        Exports the compressed model to the TensorFlow SavedModel format.

        :param save_path: The path where the model will be saved.
        """
        self._model.save(save_path, save_format=TFExportFormat.SAVED_MODEL)
        model = tf.saved_model.load(save_path)
        tf.saved_model.save(model, save_path)

    def _export_to_h5(self, save_path: str) -> None:
        """
        Exports the compressed model to the Keras H5 format.

        :param save_path: The path where the model will be saved.
        """
        self._model.save(save_path, save_format=TFExportFormat.KERAS_H5)

    def _export_to_frozen_graph(self, save_path: str) -> None:
        """
        Exports the compressed model to the Frozen Graph format.

        :param save_path: The path where the model will be saved.
        """
        # Convert Keras model to the frozen graph.
        input_signature = []
        for item in self._model.inputs:
            input_signature.append(tf.TensorSpec(item.shape, item.dtype, item.name))
        concrete_function = tf.function(self._model).get_concrete_function(input_signature)
        frozen_func = convert_variables_to_constants_v2(concrete_function, lower_control_flow=False)
        frozen_graph = frozen_func.graph.as_graph_def(add_shapes=True)

        save_dir, name = os.path.split(save_path)
        tf.io.write_graph(frozen_graph, save_dir, name, as_text=False)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import random
from itertools import combinations
from typing import Any, Dict, List, Optional

from nncf.common.graph import NNCFNodeName
from nncf.common.graph.transformations.commands import TransformationCommand
from nncf.common.logging import nncf_logger
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.base_handler import ELASTICITY_BUILDERS
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.base_handler import ELASTICITY_HANDLERS_MAP
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.base_handler import ELASTICITY_PARAMS
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.base_handler import BaseElasticityParams
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.base_handler import SingleElasticityBuilder
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.base_handler import SingleElasticityHandler
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.elastic_width import ElasticWidthHandler
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.elasticity_dim import ElasticityDim
from nncf.experimental.torch.search_building_blocks.search_blocks import BuildingBlock
from nncf.experimental.torch.search_building_blocks.search_blocks import BuildingBlocks
from nncf.experimental.torch.search_building_blocks.search_blocks import ExtendedBuildingBlocks
from nncf.experimental.torch.search_building_blocks.search_blocks import GroupedBlockIDs
from nncf.experimental.torch.search_building_blocks.search_blocks import get_building_blocks
from nncf.experimental.torch.search_building_blocks.search_blocks import get_group_of_dependent_blocks
from nncf.torch.dynamic_graph.context import TracingContext
from nncf.torch.nncf_network import NNCFNetwork

BlockId = int
ElasticDepthConfig = List[BlockId]  # list of block indexes
ElasticDepthSearchSpace = List[ElasticDepthConfig]  # grouped list of block indexes


class EDHandlerStateNames:
    DEPTH_INDICATOR = "depth_indicator"


class ElasticDepthHandler(SingleElasticityHandler):
    """
    An interface for handling elastic depth dimension in the network, i.e. skip some layers in the model.
    """

    _depth_state_names = EDHandlerStateNames

    def __init__(
        self,
        target_model: NNCFNetwork,
        skipped_blocks: BuildingBlocks,
        skip_dependencies: GroupedBlockIDs,
        node_names_per_block: Dict[int, List[NNCFNodeName]],
    ):
        """
        Constructor

        :param target_model: a target NNCFNetwork for adding modifications
        :param skipped_blocks: list of building blocks to be skipped.
        :param skip_dependencies: indexes of building blocks grouped by connectivity.
        Blocks that follow each other in the graph (i.e. they are connected by one edge) belong to the same group.
        :param node_names_per_block: mapping of block id and all node names inside the block.
        """
        super().__init__()
        self._target_model = target_model
        self._tracing_context: TracingContext = self._target_model.nncf.get_tracing_context()
        self._skipped_blocks = skipped_blocks
        self._skip_dependencies = skip_dependencies
        self._node_names_per_block = node_names_per_block
        self._depth_indicator = 1
        self._is_search_space_obsolete = True
        self._cached_search_space = None

    def get_transformation_commands(self) -> List[TransformationCommand]:
        """
        :return: transformation commands for introducing the elasticity to NNCFNetwork
        """
        return []

    @property
    def depth_indicator(self) -> int:
        """
        :return: depth indicator value that restricts number of skipped blocks in the independent groups of blocks
        """
        return self._depth_indicator

    @depth_indicator.setter
    def depth_indicator(self, depth_indicator: int) -> None:
        """
        Sets depth indicator value that restricts number of skipped blocks in the independent groups of blocks

        :param depth_indicator: depth indicator value
        """
        self._depth_indicator = depth_indicator
        self._is_search_space_obsolete = True

    def load_state(self, state: Dict[str, Any]) -> None:
        """
        Initializes object from the state.
        :param state: Output of `get_state()` method.
        """
        super().load_state(state)
        self.depth_indicator = state[self._depth_state_names.DEPTH_INDICATOR]

    def get_state(self) -> Dict[str, Any]:
        """
        Returns a dictionary with Python data structures (dict, list, tuple, str, int, float, True, False, None) that
        represents state of the object.
        :return: state of the object
        """
        state = super().get_state()
        state[self._depth_state_names.DEPTH_INDICATOR] = self.depth_indicator
        return state

    def get_search_space(self) -> ElasticDepthSearchSpace:
        """
        :return: search space that is produced by iterating over all elastic depth parameters
        """
        if not self._is_search_space_obsolete:
            return self._cached_search_space
        range_block_ids = range(0, len(self._skipped_blocks))
        # TODO(nlyalyus): can be a huge search space. no need to iterate and filter all of them. (ticket 69746)
        possible_depth_configs = [list(combinations(range_block_ids, i + 1)) for i in range_block_ids]
        possible_depth_configs = [y for x in possible_depth_configs for y in x]
        valid_depth_configs = []
        for d_sample in possible_depth_configs:
            if list(d_sample) == self._remove_inconsistent_blocks(list(d_sample)):
                valid_depth_configs.append(list(d_sample))
        self._cached_search_space = valid_depth_configs
        self._is_search_space_obsolete = False
        if [] not in valid_depth_configs:
            valid_depth_configs.append([])
        return valid_depth_configs

    def get_active_config(self) -> ElasticDepthConfig:
        """
        Forms an elasticity configuration that describes currently activated Subnet - indexes of all skipped blocks.

        :return: list of blocks' indexes to skip
        """
        return self._tracing_context.active_block_indexes

    def get_random_config(self) -> ElasticDepthConfig:
        """
        Forms an elasticity configuration that describes a Subnet with randomly chosen elastic depth

        :return: list of blocks' indexes to skip
        """
        num_blocks = len(self._tracing_context.skipped_blocks)
        config = []
        for i in range(0, num_blocks):
            is_skipped = bool(random.getrandbits(1))  # random.randint(0, 1)
            if is_skipped:
                config.append(i)
        return config

    def get_minimum_config(self) -> ElasticDepthConfig:
        """
        Forms an elasticity configuration that describes a Subnet with minimum elastic depth, i.e. maximum number of
        blocks is skipped.

        :return: list of blocks' indexes to skip
        """
        return list(range(0, len(self._tracing_context.skipped_blocks)))

    def get_maximum_config(self) -> ElasticDepthConfig:
        """
        Forms an elasticity configuration that describes a Subnet with maximum elastic depth, i.e. minimum number of
        blocks is skipped.

        :return: list of blocks' indexes to skip
        """
        return []

    def activate_supernet(self) -> None:
        """
        Activates the Supernet - the original network without skipping any blocks.
        """
        self.activate_maximum_subnet()

    def activate_subnet_for_config(self, config: ElasticDepthConfig):
        """
        Activates a Subnet that doesn't execute the blocks specified in the given elasticity configuration

        :param config: list of blocks' indexes to skip
        """
        config = self._remove_inconsistent_blocks(config)
        self._tracing_context.set_active_skipped_block(config)

    def resolve_conflicts_with_other_elasticities(
        self, config: ElasticDepthConfig, elasticity_handlers: ELASTICITY_HANDLERS_MAP
    ) -> ElasticDepthConfig:
        """
        Resolves a conflict between the given elasticity config and active elasticity configs of the given handlers.
        For example, elastic width configuration may contradict to elastic depth one. When we activate some
        configuration in the Elastic Width Handler, i.e. define number of output channels for some layers, we
        change output shapes of the layers. Consequently, it affects the blocks that can be skipped by Elastic Depth
        Handler, because input and output shapes may not be identical now.

        :param config: elasticity configuration
        :param elasticity_handlers: map of elasticity dimension to elasticity handler
        :return: elasticity configuration without conflicts with other active configs of other elasticity handlers
        """
        result = config
        if ElasticityDim.WIDTH in elasticity_handlers:
            width_handler = elasticity_handlers[ElasticityDim.WIDTH]
            assert isinstance(width_handler, ElasticWidthHandler)
            blocks = [self._tracing_context.skipped_blocks[block_idx] for block_idx in config]
            pairs_of_nodes = [(block.start_node_name, block.end_node_name) for block in blocks]
            if pairs_of_nodes:
                indexes_of_pairs = width_handler.find_pairs_of_nodes_with_different_width(pairs_of_nodes)
                if indexes_of_pairs:
                    result = [element for idx, element in enumerate(config) if idx not in indexes_of_pairs]
                    nncf_logger.debug(
                        f"The blocks with indexes {indexes_of_pairs} are not skipped to avoid inconsistency with width"
                    )
        return result

    def get_names_of_skipped_nodes(self) -> List[NNCFNodeName]:
        """
        :return: names of nodes that are currently skipped by Elastic Depth
        """
        names_of_skipped_nodes = []
        active_block_indexes = self._tracing_context.active_block_indexes
        if active_block_indexes is not None:
            for idx in active_block_indexes:
                names_of_skipped_nodes.extend(self._node_names_per_block[idx])
        return names_of_skipped_nodes

    def _remove_inconsistent_blocks(self, config: ElasticDepthConfig) -> ElasticDepthConfig:
        return self._remove_blocks_skipped_non_progressively(config)

    def _remove_blocks_skipped_non_progressively(self, config: ElasticDepthConfig) -> ElasticDepthConfig:
        assert self._skip_dependencies is not None, "Please include depth dependencies in conf. Pending automation."
        block_indexes_to_remove = []
        for block_index in config:
            tmp_block_indexes_to_remove = self._get_block_indexes_to_remove(block_index, config)
            block_indexes_to_remove.extend(tmp_block_indexes_to_remove)

        for block_index in block_indexes_to_remove:
            config.remove(block_index)
            nncf_logger.debug(f"The block #{block_index} is not skipped to not violate progressive shrinking")
        return config

    def _get_block_indexes_to_remove(self, block_index: int, config: ElasticDepthConfig) -> ElasticDepthConfig:
        block_indexes_to_remove = []
        for group in self._skip_dependencies.values():
            found = False
            group_index = group.index(block_index) if block_index in group else None
            if group_index is not None:
                found = True
                if len(group) - group_index > self.depth_indicator:
                    nncf_logger.debug(f"The block with {block_index} did not pass the depth_indicator test")
                    block_indexes_to_remove.append(block_index)
                    break
                valid_block_indexes = [group[group_index]]
                for i in range(group_index + 1, len(group)):
                    if group[i] in config:
                        valid_block_indexes.append(group[i])
                    else:
                        nncf_logger.debug(
                            f"The block #{block_index} or #{valid_block_indexes} "
                            f"did not satisfy requirement of next static block"
                        )
                        for valid_block_index in valid_block_indexes:
                            block_indexes_to_remove.append(valid_block_index)
                        break
            if found:
                break
        return block_indexes_to_remove


class EDBuilderStateNames:
    SKIPPED_BLOCKS = "skipped_blocks"
    SKIPPED_BLOCKS_DEPENDENCIES = "skipped_blocks_dependencies"


class EDParamsStateNames:
    MAX_BLOCK_SIZE = "max_block_size"
    MIN_BLOCK_SIZE = "min_block_size"
    HW_FUSED_OPS = "hw_fused_ops"
    SKIPPED_BLOCKS = "skipped_blocks"


@ELASTICITY_PARAMS.register(ElasticityDim.DEPTH)
class ElasticDepthParams(BaseElasticityParams):
    _state_names = EDParamsStateNames

    def __init__(
        self,
        max_block_size: int,
        min_block_size: int,
        hw_fused_ops: bool = True,
        skipped_blocks: Optional[List[List[NNCFNodeName]]] = None,
    ):
        """
        Constructor

        :param max_block_size: Defines minimal number of operations in the block. Option is available for the auto mode
         only. Default value is 50.
        :param min_block_size: Defines minimal number of operations in the skipping block. Option is available for the
         auto mode only. Default value is 6.
        :param hw_fused_ops: If True, automatic block search will not relate operations, which are fused on inference,
         into different blocks for skipping. True, by default.
        :param skipped_blocks: list of building blocks to be skipped. The block is defined by names of start and end
        nodes. If the parameter is not specified blocks to skip are found automatically.
        """
        self.max_block_size = max_block_size
        self.min_block_size = min_block_size
        self.hw_fused_ops = hw_fused_ops
        self.skipped_blocks = skipped_blocks

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "ElasticDepthParams":
        """
        Creates the object from its config.
        """
        kwargs = {
            cls._state_names.MAX_BLOCK_SIZE: config.get(cls._state_names.MAX_BLOCK_SIZE, 50),
            cls._state_names.MIN_BLOCK_SIZE: config.get(cls._state_names.MIN_BLOCK_SIZE, 5),
            cls._state_names.HW_FUSED_OPS: config.get(cls._state_names.HW_FUSED_OPS, True),
            cls._state_names.SKIPPED_BLOCKS: config.get(cls._state_names.SKIPPED_BLOCKS),
        }
        return cls(**kwargs)

    @classmethod
    def from_state(cls, state: Dict[str, Any]) -> "ElasticDepthParams":
        """
        Creates the object from its state.

        :param state: Output of `get_state()` method.
        """
        return cls(**state)

    def get_state(self) -> Dict[str, Any]:
        """
        Returns the compression loss state.

        :return: The compression loss state.
        """
        return {
            self._state_names.MAX_BLOCK_SIZE: self.max_block_size,
            self._state_names.MIN_BLOCK_SIZE: self.min_block_size,
            self._state_names.HW_FUSED_OPS: self.hw_fused_ops,
            self._state_names.SKIPPED_BLOCKS: self.skipped_blocks,
        }

    def __eq__(self, other: "ElasticDepthParams") -> bool:
        return self.__dict__ == other.__dict__

    def __str__(self):
        return (
            f"{self.__class__.__name__}: max_block_size: {self.max_block_size} "
            f"min_block_size: {self.min_block_size} "
            f"hw_fused_ops: {self.hw_fused_ops} skipped_blocks: {self.skipped_blocks}"
        )


@ELASTICITY_BUILDERS.register(ElasticityDim.DEPTH)
class ElasticDepthBuilder(SingleElasticityBuilder):
    """
    Determines which modifications should be made to the original FP32 model in order to introduce elastic depth
    to the model.
    """

    _state_names = EDBuilderStateNames

    def __init__(
        self,
        params: ElasticDepthParams,
        ignored_scopes: Optional[List[str]] = None,
        target_scopes: Optional[List[str]] = None,
    ):
        """
        :param params: parameters to configure elastic depth.
        :param ignored_scopes: A list of strings to match against NNCFGraph node names
          and ignore matching nodes. Ignored nodes will not considered for skipping.
        :param target_scopes: A list of strings to match against NNCFGraph and define a set of nodes
          to be considered for skipping. When `ignored_scopes` is a "denylist",
          the `target_scopes` is an "allowlist"; otherwise, same effects apply as for `ignored_scopes`.
        """
        super().__init__(ignored_scopes, target_scopes)
        self._params = params
        self._skipped_blocks: Optional[ExtendedBuildingBlocks] = None
        self._skip_dependencies: Optional[GroupedBlockIDs] = None
        if self._params.skipped_blocks is not None:
            self._skipped_blocks = [BuildingBlock(*b) for b in self._params.skipped_blocks]
            self._skip_dependencies = get_group_of_dependent_blocks(self._skipped_blocks)

    def build(self, target_model: NNCFNetwork) -> ElasticDepthHandler:
        """
        Creates modifications to the given NNCFNetwork for introducing elastic depth and creates a handler object that
        can manipulate this elasticity.

        :param target_model: a target NNCFNetwork for adding modifications
        :return: a handler object that can manipulate the elastic depth.
        """
        tracing_context = target_model.nncf.get_tracing_context()
        tracing_context.elastic_depth = True
        blocks_for_info = self._skipped_blocks
        if self._skipped_blocks is None:
            extended_blocks, self._skip_dependencies = get_building_blocks(
                target_model,
                max_block_size=self._params.max_block_size,
                min_block_size=self._params.min_block_size,
                hw_fused_ops=self._params.hw_fused_ops,
            )
            self._skipped_blocks = [eb.basic_block for eb in extended_blocks]
            blocks_for_info = extended_blocks
        nncf_logger.info("Blocks for skipping (changing the depth of model):")
        new_line = "\n"
        str_bs = [str(block) for block in blocks_for_info]
        nncf_logger.info(f"\n[{new_line.join(str_bs)}]\n\n")

        tracing_context.set_elastic_blocks(self._skipped_blocks)
        node_names_per_block = self._get_node_names_per_block(target_model, self._skipped_blocks)

        return ElasticDepthHandler(target_model, self._skipped_blocks, self._skip_dependencies, node_names_per_block)

    def load_state(self, state: Dict[str, Any]) -> None:
        """
        Initializes object from the state.

        :param state: Output of `get_state()` method.
        """
        params_from_state = state[SingleElasticityBuilder._state_names.ELASTICITY_PARAMS]
        params = ElasticDepthParams.from_state(params_from_state)
        if self._params and self._params != params:
            nncf_logger.warning(
                "Different elasticity parameters were provided in two places: on init and on loading "
                "state. The one from state is taken by ignoring the ones from init."
            )
        self._params = params
        skipped_blocks_from_state = state[self._state_names.SKIPPED_BLOCKS]
        self._skip_dependencies = state[self._state_names.SKIPPED_BLOCKS_DEPENDENCIES]
        self._skipped_blocks = [BuildingBlock.from_state(bb_state) for bb_state in skipped_blocks_from_state]

    def get_state(self) -> Dict[str, Any]:
        """
        Returns a dictionary with Python data structures (dict, list, tuple, str, int, float, True, False, None) that
        represents state of the object.

        :return: state of the object
        """
        skipped_blocks_from_state = []
        if self._skipped_blocks is not None:
            skipped_blocks_from_state = list(map(lambda x: x.get_state(), self._skipped_blocks))
        return {
            SingleElasticityBuilder._state_names.ELASTICITY_PARAMS: self._params.get_state(),
            self._state_names.SKIPPED_BLOCKS: skipped_blocks_from_state,
            self._state_names.SKIPPED_BLOCKS_DEPENDENCIES: self._skip_dependencies,
        }

    @staticmethod
    def _get_node_names_per_block(target_model: NNCFNetwork, skipped_blocks) -> Dict[int, List[NNCFNodeName]]:
        graph = target_model.nncf.get_original_graph()
        all_node_names_per_block = {}
        for idx, block in enumerate(skipped_blocks):
            simple_paths = graph.get_all_simple_paths(block.start_node_name, block.end_node_name)

            node_names_in_block = set()
            for node_keys_in_path in simple_paths:
                for node_key in node_keys_in_path:
                    node = graph.get_node_by_key(node_key)
                    node_names_in_block.add(node.node_name)

            node_names_in_block.remove(block.start_node_name)
            all_node_names_per_block[idx] = list(node_names_in_block)
        return all_node_names_per_block
